/**************************************************************************
                        VARMINT'S AUDIO TOOLS 0.6

  SOUND.C
  Written by: Eric Jorgensen (May, 1995)
                smeagol@rt66.com



  Written for: Turbo C++ V 3.1

    Compiler Settings:

     - Stack checking should be turned off
     - The default memory model for VAT is LARGE
     - register variables should be turned off
     - The setting for "assume SS == DS" should be "never".

    Linker Settings:

     - "No stack" warning should be turned off


                           **** WARNING ****

 Use Varmint's Audio Tools at your own risk.  This code has only
 been lightly tested and has been found to cause cancer in laboratory rats.

**************************************************************************/



#include "sound.h"

//-------------------------------- Local Defs


//#define CINTNO         0
//#define SLAVEPIC       2
//#define RTCINTNO       8
#define WORD           unsigned int
#define BYTE           unsigned char
#define TD             midi_data->track[i] + midi_trackloc[i]
#define DOWN_SEMITONE 1.059463
#define UP_SEMITONE   0.943874
#define DOSYNC        if(sync_on && inportb(0x3da)&0x08) vsyncclock = 1000
#define DMAAUTOINIT   0x58
#define DMAONESHOT     0x48
#define VDB           strcat(vatdebugstring,vatapstr)


//-------------------------------- Internal Function prototypes

int                   cardcheck(void);
static cardtype       checkhard(void);
void                   dmaset(BYTE far *sound_address,WORD len,BYTE channel);
void                   dodiv(void);
void                   dotick(void);
void                   enableint(BYTE nr);
void                   freeeffect(DWORD id);
void                   freesound(DWORD id);
void                   disableint(BYTE nr);
int                   getsbenv(void);
int                   getvoice(VOICE midi_voice[],int track,int channel, int note);
SOUNDEFFECT far       *grabsound(DWORD id);
BYTE                   int2vect(BYTE intnr);
WORD                   mcalc(WORD micro);
void _saveregs        mdelay(WORD delay);
void                   measure(void);
void                   measure0(void);
void                   midiplayer(void);
SOUNDEFFECT far       *newsoundeffect(void);

//WORD                   polldma(BYTE channel);
long int               readlong(FILE *infile);
int                   readshort(FILE *infile);
int                   readvarlen(BYTE *data,long int *value);
void far _saveregs _loadds interrupt sbint(void);
void                   sbremovevector(void);
void                   sbsetvector(void);
static int             scanint(void);
static int             testint(void);
void                   venable(void);
void                   vdisable(void);
cardtype               whichcard(void);


//-------------------------------- Global function pointers
                                // These are used by the interrupt routines
                                // To help keep track of interrupts that
                                // Get shuffled around.

void far _saveregs _loadds interrupt  (*handlerint)(void) = sbint;
static void far interrupt (*orgint)(void) = NULL;
static void far interrupt (*orgtick)(void)= NULL;
static void far interrupt (*orgirqint)(void) = NULL;
static void far           (*call_func)(void);

//-------------------------------- FM variables and data

static WORD FM_off[9]={0,0x100,0x200,0x800,0x900,0xa00,0x1000,0x1100,0x1200};
static BYTE FM_fnr[12]={0x57,0x6b,0x81,0x98,0xb0,0xca,0xe5,0x02,0x20,0x41,0x63,0x87};
static BYTE FM_key_or[12]={1,1,1,1,1,1,1,2,2,2,2,2};
static BYTE FM_key[9];
static BYTE FM_keyscale1[9];
static BYTE FM_keyscale2[9];
static BYTE FM_vol[9] = {0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f};

//-------------------------------- setup data and variables

static WORD ioaddr[6]={0x220,0x240,0x210,0x230,0x250,0x260};
static BYTE intrx[8]={5,7,2,3,10,11,12,15};
WORD         io_addr= 0x220;
WORD         intnr= 5;
WORD         dma_ch=1;
WORD         card_id=1;
WORD         fm_addr;
WORD         dsp_vers;
static WORD mue3,mue23,rythm=0xbd00;
static WORD dma_adr[8]= {0x00,0x02,0x04,0x06,0xc0,0xc4,0xc8,0xcc};
static WORD dma_len[8]= {0x01,0x03,0x05,0x07,0xc2,0xc6,0xca,0xce};
static WORD dma_page[8]={0x87,0x83,0x81,0x82,0x8f,0x8b,0x89,0x8a};
SBERROR     sberr = 0;
char         *errname[] = {
              "Cannot detect FMchip",
              "Cannot detect DSP",
              "Cannot find an open IRQ",
              "Cannot find an open DMA channel",
              "Cannot allocate memory for DMA buffer"};
int         tst_cnt;
BYTE         dma_controlbyte = DMAAUTOINIT;

//-------------------------------- DMA/DSP mixer varaibles and data

WORD         dma_bufferlen = 16;
BYTE far     *dma_buffer[2] = {NULL,NULL};
int         dma_currentbuffer = 0;
int far     *mix_buffer = NULL;
int         sounds_in_queue = 0;
int         sample_rate = 11000;
SAMPLE       *sounddata[MAXSOUNDS];
DWORD       soundpos[MAXSOUNDS];

SOUNDEFFECT far *soundlist = NULL;
SOUNDEFFECT far soundtoken[MAXSOUNDTOKENS];
DWORD       sampleid = 1;


//-------------------------------- Timer varaibles

static WORD timer_val,timer_hold,timer_diff,mue999;
static WORD timadd,timsum;

//-------------------------------- MOD varaibles

MOD far      *mod_data = NULL;
int         mod_on = FALSE,mod_reset = FALSE,mod_tablepos=0;
BYTE far     *mod_pattern,*mod_pstop,*mod_loopspot=NULL;
int         mod_loopcount = -1;
int         mod_bytespertick = 220,mod_ticksperdivision = 6;
int         mod_currentbyte = 0,mod_currenttick = 0;
int         mod_divoffset =0;
int         mod_glissando = FALSE;
int         mod_finetune = 0;
int         mod_patterndelay = 0;
int         mod_volume = 7;
BYTE         mod_channelselect[4] = {1,1,1,1};

CHANNEL     chan[4];



//-------------------------------- MIDI/MPU varaibles

MIDI far    *midi_data = NULL;
int         midi_reset = TRUE;
int         midi_on = FALSE;
int         midi_mpuout = FALSE;
int         midi_fmout = TRUE;
WORD         midi_mpuport = 0x330;
float       midi_callfreq = 1.0;
int           midi_usertempo = 200;
float       midi_tempoadjust = 0.5;
BYTE         midi_volume = 0x32;
BYTE         defaultpatchmap[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
BYTE         *midi_patchmap;
int         mpu_timeout;
int         mpu_available = FALSE;
VOICE        midi_voice[32];
BYTE        midi_lastevent[32];
BYTE        midi_trackon[32];
WORD        midi_trackloc[32];
float       midi_timer[32];
int         midi_divisions = 96;
int          midi_msperdiv=5000;
int          midi_livetracks;
long int     midi_count = 100;


//-------------------------------- Miscellaneous varaibles

int         debugnum=0;
int         debug_lowdsp = FALSE;
int         debug_reverseflipflop = FALSE;
int         debug_forceflipflop = -1;
int         disabled = 0;
WORD         DSP_overhead = 0;

extern unsigned _stklen = 16000;  // The stack is usually only 4K, but
                                  // I was getting stack overflow problems
                                  // with that ammount.

WORD vsync_toolong = 4;           // This is the number of vclock ticks that
                                  // tells us we have waited to long in
                                  // the VarmintVSync() function.
DWORD       vclock=0;
WORD         vsyncclock = 0;
BYTE         sync_on = FALSE;
                                  // This string is here to help debug
                                  // startup problems
char        vatdebugstring[5000] = {"*** DEBUGGING STRING ***\n"};
char        vatapstr[255];



/**************************************************************************
  void GoVarmint(void)
  void DropDeadVarmint(void)

  DESCRIPTION:  Starts/stops the interrupt routine for Varmint's audio tools

**************************************************************************/
void GoVarmint(void)
{
  sbsetvector();                           // Install the sound kernel

  if(debug_lowdsp) dsp_vers = 0;

                                           // Seems like later DSP versions
                                          // need to have the flipflop in
                                          // reverse order to remove static
  if(dsp_vers >= 0x300) debug_reverseflipflop = TRUE;
  if(debug_forceflipflop != -1) debug_reverseflipflop = debug_forceflipflop;


  if(dsp_vers >= 512) {                   // SB2.0 or better?
    DSPWrite(0x48);                       // Set DSP for 8bit Autoinit DMA
    DSPWrite((dma_bufferlen-1) & 0xff);    // Write length low byte
    DSPWrite((dma_bufferlen-1) >> 8);      // Write length high byte

    dmaset(dma_buffer[0],dma_bufferlen*2-1,dma_ch);
    DSPWrite(0x1C);                       // Set DSP for Autoinit 8bit DMA
  }
  else {                                  // SB 1.x?
    dma_controlbyte = DMAONESHOT;
  }

  DSPWrite(DSP_INVOKE_INTR);               // Ignition!
}

void DropDeadVarmint(void)
{
  DSPWrite(0xD0);                          // Halt DMA
  if(dsp_vers >= 512) DSPWrite(0xDA);     // Halt Autoinitialized DMA
  DSPWrite(0xD0);                          // Halt DMA
  sbremovevector();                        // Clean up sound kernel
}

/*  ---------------  FM  Stuff ------------ */



/**************************************************************************
  void FMWrite(WORD data)

  DESCRIPTION: Writes a byte to the FM chip.

  INPUTS:
      data    High byte contains the register number, low byte
              contains the data to write to that register.

**************************************************************************/
void FMWrite(WORD data)
{
  BYTE reg,value;

  reg = (data & 0xff00)>>8;       // extract register and data value
  value = data & 0x00ff;

  mdelay(mue3);                   // Wait three microseconds
  outportb(fm_addr,reg);          // Write the register
  mdelay(mue3);                   // Wait three microseconds

  outportb(fm_addr+1,value);      // Write the data
  mdelay(mue23);                  // Wait 23 microseconds

}


/**************************************************************************
  void FMReset()

  DESCRIPTION:  Resets the FM chip by clearing all the registers then
                setting a few appropriate bits.

**************************************************************************/
void FMReset(void)
{
  WORD i;
  for(i = 0; i <= 0xf500 ; i+= 0x100) FMWrite(i);
  FMWrite(0x0120);                 // Turn on Wave form control
  FMWrite(0xbdc0);                 // Set AM and Vibrato to high
}


/**************************************************************************
  BYTE FMStatus()

  DESCRIPTION:  Reads the status byte of the FM chip

  RETURNS:  The value at fm_addr

**************************************************************************/
BYTE FMStatus(void)
{
  return (inportb(fm_addr));

}


/**************************************************************************
  int FMDetect()

  DESCRIPTION:  Detects the presence of an FM chip

  RETURNS:  A boolean value indicating wether or not test was successful

**************************************************************************/
int FMDetect(void)
{
  FMWrite(0x0100); /* init Test register */

  FMWrite(0x0460); /* reset both timer */
  FMWrite(0x0480); /* enable interrupts */
  if(FMStatus() & 0xe0) return(FALSE);

  FMWrite(0x02ff); /* write ffh to timer 1 */
  FMWrite(0x0421); /* start timer 1 */
  if(fm_addr==0x388) MilliDelay(42); /* wait at least 21000 mcs */
  else mdelay(mcalc(160));   /* wait at least 80 microsec */
  if((FMStatus() & 0xe0)!=0xc0) return(FALSE);

  FMWrite(0x0460); /* reset both timer */
  FMWrite(0x0480); /* enable interrupts */
  return(TRUE);
}



/**************************************************************************
  void FMSetVoice(BYTE voice,BYTE *ins)

  DESCRIPTION: Sets the voice from an 11 byte array

  INPUTS:

    voice    FM voice number (0-8)
    ins      A pointer to an array of voice descriptor  bytes

    BYTE    ID

      0      Ampmod /vib /envtype /scale rate/ mod freq mult (oper 1)
      1      Ampmod /vib /envtype /scale rate/ mod freq mult (oper 2)
      2      Key level scaling/ total level (oper 1)
      3      Key level scaling/ total level (oper 2)
      4      Attack Rate/ Decay rate  (oper 1)
      5      Attack Rate/ Decay rate  (oper 2)
      6      Sustain Level/ Release rate (oper 1)
      7      Sustain Level/ Release rate (oper 2)
      8     Feedback / Algorythm (oper 1&2)
      9      Wave Form  Select (oper 1)
      10    Wave Form  Select (oper 2)

**************************************************************************/
void FMSetVoice(BYTE voice,BYTE *ins)
{
  if(voice > 8) return;

  FM_keyscale1[voice]=ins[2] & 0xc0;        // store key scaling for FM_Vol
  FM_keyscale2[voice]=ins[3] & 0xc0;
                                            // Write voice data
  FMWrite((0x2000 + FM_off[voice]) | ins[0]);
  FMWrite((0x2300 + FM_off[voice]) | ins[1]);
                                            // For the next two, we want to
                                            // make sure current volume is
                                            // preserved.
  FMWrite((0x4000 + FM_off[voice]) | (ins[2] & 0xc0) | FM_vol[voice]);
  FMWrite((0x4300 + FM_off[voice]) | (ins[3] & 0xc0) | FM_vol[voice]);
                                            // the rest of the voice is just
                                            // straight writes.
  FMWrite((0x6000 + FM_off[voice]) | ins[4]);
  FMWrite((0x6300 + FM_off[voice]) | ins[5]);
  FMWrite((0x8000 + FM_off[voice]) | ins[6]);
  FMWrite((0x8300 + FM_off[voice]) | ins[7]);
  FMWrite((0xc000 + voice * 0x100) | ins[8]);
  FMWrite((0xE000 + FM_off[voice]) | ins[9]);
  FMWrite((0xE300 + FM_off[voice]) | ins[10]);
}


/**************************************************************************
  void FMSetFrequency(BYTE voice,int freq)

  DESCRIPTION: sets an explicit pseudo frequency (0 - 0xffff)

  INPUTS:

    voice    FM voice (0-8)
    freq    number from 0 to FFFF. This does not really correspond
            to the output frequency other than low values put
            out low frequencies and vice versa.

  Note: There is no way to really set a direct frequency on an FM
        chip, so I wrote this routine which is based on octaves. I
        imagine it is slightly non-linear.  Still, it is good for
        special effects.

**************************************************************************/
void FMSetFrequency(BYTE voice,WORD freq)
{
  BYTE highbits,lowbits;
  WORD data,frac;
  int octave;

  octave = (freq / 0x2000);       // Extract octtave number (0-7)
                                  // convert remaining fraction into a
                                  // 10 bit value.
  frac = ((double)(freq - octave * 0x2000)/(double)0x2000) * 0x157 + 0x157;

  highbits = (frac & 0x300) >> 8; // divide fraction into low and high bits
  lowbits = frac & 0xff;

  data=0xa000+(voice<<8)|lowbits; // store low bits for now
  FM_key[voice]=highbits|(octave<<2);  // save high bits for Key_on(); (octave 4)
  FMWrite(data);                 // write low bits to FM chip;
}

/**************************************************************************
  void FMSetNote(BYTE voice,BYTE note)

  DESCRIPTION: sets the frequency for a chromatic note

  INPUTS:
    voice    FM voice #  (0-8)
    note    Regular MIDI note (0-127)

**************************************************************************/
void FMSetNote(BYTE voice,BYTE note)
{
  BYTE blk,notex;
  WORD data;
                              //  calculate freq number and octave
  notex=note-24;
  blk=1;
  while(notex>=12)
  {
    notex-=12;
    blk++;                    // octave number
  }

  data=0xa000+(voice<<8)|FM_fnr[notex];
  FM_key[voice]=FM_key_or[notex]|(blk<<2); // save part of the note for Key_on()
  FMWrite(data);             // write note to the chip
}


/**************************************************************************
  void FMSetVolume(BYTE voice,BYTE vol)

  DESCRIPTION: The the volume (0-63) for a voice.

  INPUTS:
    voice    FM voice #  (0-8)
    vol      volume value (0-63)

**************************************************************************/
void FMSetVolume(BYTE voice,BYTE vol)
{
  if (voice >8) return;
                               // Convert  volume from a logical value to
                               // the value that is really used.  ie: 3f is
                               // really the quietest setting, while 0 is
                               // the loudest.  Weird, eh?
  FM_vol[voice] = (0x3f - (vol & 0x3f));
                               // Write the volume while preserving the
                               // other important parts of the voice.
  FMWrite((0x4000+FM_off[voice]) |FM_vol[voice] | FM_keyscale1[voice]);
  FMWrite((0x4300+FM_off[voice]) |FM_vol[voice] | FM_keyscale2[voice]);

}

/*  A NOTE ABOUT RYTHM FUNCTIONS:

  I've only played around with these functions a little bit. Here are some
  things that I've learned:

    - only channels 6,7,and 8 are affected by the rythm mode.
    - You will need to develop special instrument definitions to get
      the rythm instruments to sound right.  The most important parameters
      in a rythm instrument definition are attack/decay/sustain rates and
      the waveform (bytes 9 and 10).
    - channels 6,7, and 8 each behave differently in rythm mode:

        6 - Instrumental.  Sounds like a triangle
        7 - White noise.  Sounds like a snare drum
        8 - High white noise.  Sounds like a Cymbal.

    - If you want to add white noise effects to your program (Gun shots
      engines, etc...)  channel 7 in rythm  mode is a good source.

                                      - ERIC
*/



/**************************************************************************
  void FMSetRythmMode(BYTE bool)

  DESCRIPTION:  Turns on/off rythm mode based on input.

  INPUTS:
    bool    TRUE or FALSE value for setting the rythm mode

**************************************************************************/
void FMSetRythmMode(BYTE bool)
{
  WORD data;

  if(bool) data=0xbde0;               // Set the rythm mode bit
  else data=0xbdc0;

  rythm=data;                         // This global keeps track of the
                                      // mode for other rythm functions.
  FMWrite(data);
}


/**************************************************************************
  void FMRythmOn(BYTE inst)

  DESCRIPTION: Turns on a Specified  rythm instrument.

  INPUTS:
    inst    FM rythm instrument specification.

            You should use these definitions:

                      FM_HIHAT
                      FM_TOPCYM
                      FM_TOMTOM
                      FM_SNARE
                      FM_BASS

**************************************************************************/
void FMRythmOn(BYTE inst)
{
  rythm|=inst;
  FMWrite(rythm);
}

/**************************************************************************
  void FMRythmOff(BYTE inst)

  DESCRIPTION: Turns off a Specified  rythm instrument.

  INPUTS:
    inst    FM rythm instrument specification.

            You should use these definitions:

                      FM_HIHAT
                      FM_TOPCYM
                      FM_TOMTOM
                      FM_SNARE
                      FM_BASS

**************************************************************************/
void FMRythmOff(BYTE inst)
{
  rythm&=(~inst);
  FMWrite(rythm);
}


/**************************************************************************
  void FMKeyOn(BYTE voice)

  DESCRIPTION: Turn on an FM voice.

               This description is misleading, since in my experirnce,
               FM voices are always on.  This function really just
               triggers the FM voice.

  INPUTS:
    voice    FM voice #  (0-8)

**************************************************************************/
void FMKeyOn(BYTE voice)
{
  WORD data;

  if(voice > 8) return;

  data=0xb000+(voice<<8);            // set write address
  data |= FM_key[voice]|0x20;        // set key on bit and frequency
   FMWrite(data);
}


/**************************************************************************
  void FMKeyOff(BYTE voice)

  DESCRIPTION: Turn off an FM voice.

                (See FMKeyOn)  Again, I've found that voices are always
                on, and to turn them off you really need to just set the
                volume to 0.  Turning off the Key_on bit may prepare
                the voice for a trigger, though.

  INPUTS:
    voice    FM voice #  (0-8)

**************************************************************************/
void FMKeyOff(BYTE voice)
{
  WORD data;

  if(voice > 8) return;

  data=0xb000+(voice<<8);            // set address
  data |= FM_key[voice];              // preserve frequency data
  FMWrite(data);
                                     //  working.
}



/*  ---------------  DSP  Stuff ------------ */




/**************************************************************************
  int DSPReset()

  DESCRIPTION: Resets the DSP

  RETURNS:
    TRUE or FALSE to indicate wether reset was sucsessful

**************************************************************************/
int DSPReset(void)
{
  int i;

  sprintf(vatapstr,"DSPReset() - <entry>\n"); VDB;

  mdelay(mue3);                         // Wait 3 microseconds
  outportb(io_addr+DSP_RESET,1);        // Write a 1 to the DSP reset port
  mdelay(mue3);                         // Wait 3 microseconds
  outportb(io_addr+DSP_RESET,0);        // Write a 0 to the DSP reset port


  for(i=0;i<50;i++) {                   // DSP should send back an 0xaa
    mdelay(mue3);
    if(DSPRead()==0xaa) return(TRUE);
  }

  return(FALSE);
}


/**************************************************************************
  BYTE DSPRead()

  DESCRIPTION:  reads a byte from the dsp

  RETURNS:
    The byte that was read

**************************************************************************/
BYTE DSPRead(void)
{
                                      // Read until high bit of status port
                                      // is set.
  while(!(inportb(io_addr+DSP_RSTATUS) &  0x80));
  return (inportb(io_addr+DSP_READ)); // Send back the value of the read port
}

/**************************************************************************
  void DSPWrite(BYTE output)

  DESCRIPTION: Writes a byte to the DSP

  INPUT:
    output  byte to write to the DSP

**************************************************************************/
void DSPWrite(BYTE output)
{
                                      // Read until high bit of status port
                                      // is clear.
  while((inportb(io_addr+DSP_WSTATUS) &  0x80));
  outportb(io_addr+DSP_WRITE,output); // Write our byte
}



/**************************************************************************
  WORD DSPGetVersion()

  DESCRIPTION:  Get the version number of the DSP

  RETURNS:
    DSP version number. High byte = major version, low byte = minor version

**************************************************************************/
WORD DSPGetVersion(void)
{
  DSPWrite(DSP_GET_VERS);
  return((WORD)DSPRead()*256+DSPRead());
}


/**************************************************************************
  int MPUEnter(void)

  DESCRIPTION:  This function resets the MPU-401 chip and starts UART
                transfer mode.

  RETURNS:

    True if successful, False if not

**************************************************************************/
int MPUEnter(void)
{
                                           // Check command ready bit
  for(mpu_timeout = 0; mpu_timeout < TIMEOUT; mpu_timeout++) {
    if(!(inportb(midi_mpuport) & 0x80)) break;
  }
  sprintf(vatapstr,"MPUENTER() - check ready t.o.: %u\n",mpu_timeout); VDB;

                                          //  If it never resets, it probably
                                          //  is not there.
  if(mpu_timeout == TIMEOUT) return(FALSE);

  outportb(midi_mpuport+1,0xFF);          // Send reset command

  mpu_timeout = 0;
  while(mpu_timeout<TIMEOUT) {            // Wait for ready bit to clear
    mpu_timeout++;
    if(!(inportb(midi_mpuport+1)&0x80)) break;
  }

  sprintf(vatapstr,"MPUENTER() - wait ready clear t.o.: %u\n",mpu_timeout); VDB;

  if(mpu_timeout == TIMEOUT) {             // Did it Time out?
    outportb(midi_mpuport+1,0xFF);        // Try again

    mpu_timeout = 0;
    while(mpu_timeout<TIMEOUT) {           // Wait for ready bit to clear
      mpu_timeout++;
      if(!(inportb(midi_mpuport+1)&0x80)) break;
    }
    sprintf(vatapstr,"MPUENTER() - 2nd check t.o.: %u\n",mpu_timeout); VDB;

    if(mpu_timeout == TIMEOUT) return(FALSE); // Two timeouts and you're out
  }

                                          // Check read ready bit
  for(mpu_timeout = 0; mpu_timeout < TIMEOUT; mpu_timeout++) {
    if(!(inportb(midi_mpuport) & 0x40)) break;
  }
  sprintf(vatapstr,"MPUENTER() - check read ready t.o.: %u\n",mpu_timeout); VDB;
                                          // problems?
  if(mpu_timeout == TIMEOUT) return(FALSE);

                                          // Clear out ACK byte;
  inportb(midi_mpuport);

  outportb(midi_mpuport+1,0x3F);          // Put MPU in Uart Mode

  mpu_timeout = 0;
  while(mpu_timeout<TIMEOUT) {            // Wait for ready bit to clear
    mpu_timeout++;
    if(!(inportb(midi_mpuport+1)&0x80)) break;
  }
  sprintf(vatapstr,"MPUENTER() - uart write t.o.: %u\n",mpu_timeout); VDB;

  if(mpu_timeout == TIMEOUT) return(FALSE);// No UART?  No MIDI.

  mpu_available = TRUE;
  return(TRUE);
}


/**************************************************************************
  void MPUExit(void)

  DESCRIPTION:  Sends the UART mode command to the MPU.  You must call
                MPUEnter first.

**************************************************************************/
void MPUExit(void)
{
  if(!mpu_available) return;

  outportb(midi_mpuport+1,0xFF);          // Reset the MPU
}

/**************************************************************************
  void MPUWrite(BYTE b)

  DESCRIPTION:  Writes a byte  the midi data port

  INPUTS:
    b    byte to write to the MPU-401 chip

**************************************************************************/
void MPUWrite(BYTE b)
{
  if(!mpu_available) return;

  mpu_timeout = 0;

  while(mpu_timeout<TIMEOUT) {        // Wait for ready bit to clear
    mpu_timeout++;
    if(!(inportb(midi_mpuport+1)&0x40)) break;
  }

  outportb(midi_mpuport,b);           // Write the byte!
}


/**************************************************************************
  int getsbenv()

  DESCRIPTION:  Get sound blaster information from the environment
                variable "BLASTER"

  RETURNS:
    TRUE if successful, FALSE if there was a problem

**************************************************************************/
int getsbenv(void)
{
  static char *envstr;                    // Environment strings must be
                                          // static or global
  char str[255];
  int i;
  int outvalue = TRUE;

  sprintf(vatapstr,"getsbenv() - <entry>\n"); VDB;

  envstr=getenv("BLASTER");
  if(!envstr) return(FALSE);              // no blaster variable? go home
  strcpy(str,envstr);

  sprintf(vatapstr,"getsbenv() - Raw BLASTER=%s\n",str); VDB;

                                          // Convert string to upper case
  for(i = 0 ; i < strlen(str); i++) *(str+i) = toupper(*(str+i));
                                          // pick apart variable for info.
                                          // Io address
  for(i = 0; *(str+i) != 0 && *(str + i) != 'A'; i++);
  if(*(str+i)){
    sscanf(str+i+1,"%x",&io_addr);
    if(io_addr<0x210 || io_addr>0x260) outvalue = FALSE;
  }
                                          // MIDI port address
  for(i = 0; *(str+i) != 0 && *(str + i) != 'P'; i++);
  if(*(str+i)) sscanf(str+i+1,"%x",&midi_mpuport);

                                          // Dma channel number
  for(i = 0; *(str+i) != 0 && *(str + i) != 'D'; i++);
  if(*(str+i)){
    sscanf(str+i+1,"%d",&dma_ch);
    if(dma_ch > 7) {
      outvalue = (FALSE);                  // only 0-7 allowed
      dma_ch = 1;                         // Set a reasonable default
    }
  }

                                          // IRQ interrupt number
  for(i = 0; *(str+i) != 0 && *(str + i) != 'I'; i++);
  if(*(str+i)){
    sscanf(str+i+1,"%d",&intnr);
    if(intnr < 2  || intnr > 15) {
      outvalue =  (FALSE);
      intnr = 7;
    }
  }

                                          // card_id
  for(i = 0; *(str+i) != 0 && *(str + i) != 'T'; i++);
  if(*(str+i)){
    sscanf(str+i+1,"%d",&card_id);
    if(card_id < 1 && card_id > 6) outvalue = FALSE; // 1 = SB 1.x, 6 = SB AWE
  }

  sprintf(vatapstr,"getsbenv() - Processed BLASTER=A%X P%X I%u D%u T%u\n",io_addr,midi_mpuport,intnr,dma_ch,card_id); VDB;

  return(outvalue);
}



/*  ---------------  Misc.  Stuff ------------ */



/**************************************************************************
  int cardcheck()

  DESCRIPTION:  Check for both FM chip and DSP

  RETURNS:
    Returns a value that has FM and DSP status information

**************************************************************************/
int cardcheck(void)
{
  int ret=0;

  if(FMDetect()) ret|=FM_DETECT;
  if(DSPReset()) ret|=DSP_DETECT;
  return(ret);
}


/**************************************************************************
static void far _saveregs interrupt testnint()

  DESCRIPTION:  This function is stored as an interrupt to test
                various interrupt vectors by testint()

**************************************************************************/
static void far _saveregs interrupt testnint(void)
{
  tst_cnt++;                               // increment our test counter

  inportb(io_addr+DSP_RSTATUS);            // Acknowledge DSP interrupt

  outportb(0x20,0x20);                      // Clear PIC
  outportb(0xa0,0x20);
}



/**************************************************************************
  static int testint()

  DESCRIPTION:  This function is used by scanint() to test interrupt
                stuff.  It installs a test interrupt in the
                requested spot (intnr) then sees if the DSP can
                use it.

  RETURNS:
    TRUE if successful, FALSE if not

**************************************************************************/
static int testint(void)
{
  int i;
  BYTE int1,int2;

  orgint=_dos_getvect(int2vect(intnr));     // Save original interrupt
  tst_cnt=0;                                 // reset our test interrupt counter.

  int1 = inportb(0x21);                      // Save PIC settings
  int2 = inportb(0xa1);

  vdisable();                               // Disable interrupts
  outportb(0x21,0xff);                      // Turn off all interrupts
  outportb(0xa1,0xff);

                                             // put in our test interrupt
  _dos_setvect(int2vect(intnr),testnint);
  enableint(intnr);                         //  Turn on that interrupt
  venable();                                // Enable interrupts

  DSPWrite(DSP_INVOKE_INTR);                // Force DSP interrupt

  for(i=0;i<30000;i++) if(tst_cnt) break;   // wait for interrupt code to happen

  if(i == 30000) {                          // normalize things if the interrupt
                                            // did not get called.
    inportb(io_addr+DSP_RSTATUS);           // Acknowledge DSP interrupt

    outportb(0x20,0x20);                    // Clear PIC
    outportb(0xa0,0x20);
  }

  vdisable();                               // disable inerrupts
  disableint(intnr);                         // put original interrupt back
  _dos_setvect(int2vect(intnr),orgint);

  outportb(0x21,int1);                       // restore PIC settings
  outportb(0xa1,int2);
  venable();                                // enable interrupts

  if(i==30000) return(FALSE);               // Timed out? No good!
  else return(TRUE);
}


/**************************************************************************
  static int scanint()

  DESCRIPTION: This makes sure that the interrupt number picked by the
                IRQ specification is a good choice.

  RETURNS:
    0 if failed, Interrupt number if successful
**************************************************************************/
static int scanint(void)
{
  int i;

  sprintf(vatapstr,"scanint() - <entry>\n"); VDB;

  if(testint()) return(intnr);   // Original choice good?

  sprintf(vatapstr,"scanint() - original interrupt failed (%d)\n",intnr); VDB;

  for(i=0;i<8;i++)                // Try our eight best guesses
  {
    intnr=intrx[i];
    sprintf(vatapstr,"scanint() - trying interrupt: %d\n",intnr); VDB;
    if(testint()) return(intnr);
  }

  sprintf(vatapstr,"scanint() - All interrupt tries failed\n"); VDB;

  return(0);
}


/**************************************************************************
  static cardtype checkhard()

  DESCRIPTION:  Checks hardware for DSP and FM chip

  RETURNS:
    The  detected Card Type.  (For now this is only an SB2.0)

**************************************************************************/
static cardtype checkhard(void)
{
  int ret;

  sprintf(vatapstr,"checkhard() - <entry>\n"); VDB;

  ret=DSPReset();

  sprintf(vatapstr,"checkhard() - DSPreset: %d\n",ret); VDB;

  if(ret)
  {
    if(!scanint()) {                    // Scan IRQ's
      sberr= irqerr;
      return(none);
    }
    sprintf(vatapstr,"checkhard() - scanint chose: %d\n",intnr); VDB;

    fm_addr=FM_ADLIB_ADDRESS;           // Check adlib chip first
    if(!FMDetect()) {
      fm_addr=io_addr+FM_BOTH_OFF;      // Check SB chip
      if(!FMDetect()) {
        sberr = fmerr;
        return(none);                   // no fm? -> damaged!
      }
    }
    sprintf(vatapstr,"checkhard() - FM address: %X\n",fm_addr); VDB;

    return(sb20);
  }
  sberr = nodsperr;
  return(none);
}


/**************************************************************************
  cardtype whichcard()

  DESCRIPTION:  Calls various functions to make sure you've
                got a Sound Blaster.  Also checks to make sure that
                it has the right sound blaster environment settings.

  RETURNS:
    0 on failure, cardtype on success
**************************************************************************/
cardtype whichcard(void)
{
  cardtype cret = nodsp;
  int i;

  sprintf(vatapstr,"whichcard() - <entry>\n"); VDB;

  if(getsbenv()) cret=checkhard();      // grab environment variable
  if(cret!=nodsp) return(cret);         // If dsp is there, then go home

  sprintf(vatapstr,"whichcard() - Original settings failed\n"); VDB;


  for(i=0;i<6;i++)                      // scan around for a better io address
  {
    io_addr=ioaddr[i];
    sprintf(vatapstr,"whichcard() - Trying new io address: %X\n",io_addr); VDB;

    cret=checkhard();
    if(cret!=nodsp) return(cret);
  }
  sprintf(vatapstr,"whichcard() - FAILED COMPLETELY\n"); VDB;

  return(none);                         // Uh oh.
}


/**************************************************************************
  void SBCleanUp()

  DESCRIPTION: This conveniently cleans up all the stuff done by SBSetUp()
               and other sound functions

  RETURNS:
    TRUE on success, FALSE on failure

**************************************************************************/
void SBCleanUp(void)
{
  DSPWrite(DSP_SPKR_OFF);

  farfree(dma_buffer[0]);
  farfree(mix_buffer);

  MPUExit();
}
/**************************************************************************
  int SBSetUp()

  DESCRIPTION: Sets up the sound blaster for action.  This is the only
               function a programmer should really use.  Most of the
               nitty gritty is handled internally.

  RETURNS:
    TRUE on success, FALSE on failure

**************************************************************************/
int SBSetUp(void)
{
  int i;

  InitTimerFunctions();         // Initialize counter 2 for timing functions

  sprintf(vatapstr,"SBSetUp() - InitTimerFunctions\n"); VDB;
  sprintf(vatapstr,"SBSetUp() - measure [tval: %u] [ms: %u]\n",timer_val,mue999); VDB;

                                // Calculate some special delays.  Normally
                                // These delays would be 3 and 23
                                // microseconds, but I make them twice
                                // as long just to be certain.
  mue3=mcalc(6) ;
  mue23=mcalc(46) ;

  sprintf(vatapstr,"SBSetUp() - mcalcs- [m3: %u] [m23: %u]\n",mue3,mue23); VDB;

                                        // Go and check the hardware

  if(whichcard()==none) return(FALSE);

                                         // Get DSP ready
  dsp_vers=DSPGetVersion();
  DSPWrite(DSP_SPKR_ON);

  sprintf(vatapstr,"SBSetUp() - DSP version: %d.%d\n",dsp_vers>>8,dsp_vers&0xff); VDB;

                                        // Allocate space for playback buffer
  dma_buffer[0] = (BYTE far *)farmalloc(dma_bufferlen*2+5);
  if(!dma_buffer[0]) {
    sberr = nomem;
    return(FALSE);
  }
  dma_buffer[1] = dma_buffer[0]+dma_bufferlen;

                                        // Allocate space for Mixing buffer
  mix_buffer = (int far *)farmalloc(dma_bufferlen*2+10);
  if(!mix_buffer) {
    sberr = nomem;
    return(FALSE);
  }
                                        // Clear the buffer
  for(i = 0; i <= dma_bufferlen; i++) {
    *(dma_buffer[0]+i) = 127;
    *(dma_buffer[1]+i) = 127;
  }
  SetSampleRate(11000);                 // Set the sample rate

  midi_patchmap = defaultpatchmap;


  for(i = 0; i < MAXSOUNDTOKENS; i++){  // Set up sound tokens
    soundtoken[i].type = v_unused;
  }

  sprintf(vatapstr,"SBSetUp() - Processed BLASTER=A%X P%X I%u D%u T%u\n",io_addr,midi_mpuport,intnr,dma_ch,card_id); VDB;

  return(TRUE);
}


/**************************************************************************
  DWORD far2long(char far *adr)

  DESCRIPTION: This is used by dmaset to convert a regular far address
               to a 20 bit flat address.

  RETURNS:
    The converted flat address
**************************************************************************/
DWORD far2long(char far *adr)
{
  return(((DWORD)FP_SEG(adr)<<4)+FP_OFF(adr));
}


/**************************************************************************
  void SetSampleRate(WORD rate)

  DESCRIPTION:  Sets the sample rate (specified in hz).

                Note:  The sample rate most likely will not be exactly
                equal to the rate specified in the argument argument since
                the Sound blaster can only accept rates in increments of
                400 hz.

  INPUTS:
    rate    playback rate in hertz

  RETURNS:
    The actual value of the new sample rate.

**************************************************************************/
WORD SetSampleRate(WORD rate)
{
  DWORD val;

  if(rate<4000) rate = 4000;          // 4000 is as slow as the SB can go
                                      // Calculate number for the sound card
  val=256-1000000L/rate;
  DSPWrite(DSP_SAMPLE_RATE);
  DSPWrite((BYTE)val);
                                       // This helps other functions
                                      // To know how fast the DSP is going.
  return(sample_rate = (int)(1000000L/(256-val)));
}

/**************************************************************************
  void dmaset(BYTE far *sound_address,WORD len,BYTE channel)

  DESCRIPTION:  This programs the DMA controller to start a single pass
                output transfer.

                (Draeden of VLA has provided some good information for
                DMA programming in his INTRO to DMA document)

  INPUTS:
    sound_address    Regular address of the DMA playback buffer
    len              length of the DMA buffer in bytes minus one
    channel          DMA  channel to use

**************************************************************************/
void dmaset(BYTE far *sound_address,WORD len,BYTE channel)
{
  WORD adr;
  DWORD adrl;
  BYTE page;

  adrl = far2long(sound_address);     // convert address to 20 bit format
  adr=(WORD)adrl;                     // extract page address
  page=(BYTE)(adrl>>16);              // extract page number

                                      // PREPARE DMA.
                                      //   (Channels 0-3 have different
                                      //   command ports than 4-7.)

                                      // SET CHANNEL, MASK BIT AND MODE

  if(channel <4) {                    // Chanels 0-3?
    outportb(0x0a,channel+4);          // write channel number with 3rd bit set
    outportb(0x0c,0);                  // Clear Byte Pointer

                                      // Set the mode.  The mode is determined
                                      // in GoVarmint().
    outportb(0x0b,dma_controlbyte+channel);
  }
  else {                              // channels 4-7
    outportb(0xd4,channel);            // write channel number
    outportb(0xd8,0);                  // Clear Byte Pointer

                                      // Set the mode.  The mode is determined
                                      // in GoVarmint().
    outportb(0xd6,dma_controlbyte+channel-4);
  }
                                      // SET TRANSFER INFORMATION

  outportb(dma_adr[channel],adr&0xff);// Write address low byte
  outportb(dma_adr[channel],adr>>8);  // Write address high byte

  outportb(dma_len[channel],len&0xff);// Write length  low byte
  outportb(dma_len[channel],len>>8);  // Write length  high byte

  outportb(dma_page[channel],page);   // Write page

                                      // CLEAR MASK BITS
  if(channel < 4) outportb(0x0a,channel);
  else outportb(0xd4,channel & 0x03);

}

/**************************************************************************
  void polldma(BYTE channel)

  DESCRIPTION:  This function poles the DMA controller to find out how many
                bytes are left in the current transfer.

                As of version 0.4, this function is no longer used, but
                I thought it might be useful to someone else, so I've only
                commented it out.
**************************************************************************/
/*WORD polldma(BYTE channel)
{
  BYTE low1,high1,low2,high2;

  disableint(intnr);             // Turn off the interrupt so we don't get
                                  // caught with our pants down.
  asm {
    mov dx,0x0c                   // Flip the master reset switch
    mov al,0
    out dx,al
  }
  _DX = dma_len[channel];         // Load in the counter address
                                  // read position twice, becasue sometimes
                                  // there is a problem
  asm{
    in al,dx                      // read the low byte first
    mov low1,al
    in al,dx                      // read the high byte next
    mov high1,al

    in al,dx                      // read the low byte first
    mov low2,al
    in al,dx                      // read the high byte next
    mov high2,al

  }
  enableint(intnr);              // Done, so we'll put the interrupt back.

                                  // High bytes the same? Use second reading
  if(high1 == high2)return((WORD)high2*256+low2);
                                  // else the First reading is accurate
  return(high1*256+low1);

}*/



/**************************************************************************
  void far _saveregs interrupt sbint()

  DESCRIPTION:  This is the sound Blaster interrupt that is to be
                called at the end of DMA transfer.  This is how the flow
                of things goes:

                  Mix regular samples
                  Mix MOD stuff
                  Calculate next DMA half of buffer
                  acknowledge interrupt
                  Run the midi stuff.

                The DOSYNC's are macros define in sound.h designed to
                check the vertical retrace and then set vsyncclock so that
                these sound functions can be used with games that check
                the vertical retrace.

**************************************************************************/
void far _saveregs _loadds interrupt sbint(void)
{
  int i,j;
  WORD lastbyte;
  int far *mbuf;
  BYTE far *dbuf;
  SOUNDEFFECT far *s,*hold;

  vdisable();                      // disable interrupts();

  if(DSP_overhead) TimerOn();      // start timer to measure DSP overhead

  DOSYNC;                          // Macro for checking the vertical retrace
                                   // clear out the mixing buffer
  for(j = 0; j < dma_bufferlen+1; j++) *(mix_buffer+j) = 128;
  DOSYNC;
                                   // ***********************************
                                   // *                                 *
                                   // *       Mix in plain sounds       *
                                   // *                                 *
                                   // ***********************************
  s = soundlist;
  while(s){
    hold = s->next;
    if(s->active && s->type == v_plain) {
      if(*s->pos >= dma_bufferlen) {// enough left to fill the buffer?
        for(j = 0; j < dma_bufferlen; j++) *(mix_buffer+j) += *(s->data++);
        *s->pos -= (dma_bufferlen);
      }
      else {                         // else play what is left
        for(j = 0; j <= *s->pos; j++) *(mix_buffer+j) += *(s->data++);

        if(s->chain) {               // Is there a sound to play after this?
          (s->chain)->next = soundlist;
          soundlist = s->chain;
          s->chain = NULL;           // Break link to hide from freesound()
          sounds_in_queue++;
        }
        freesound(s->id);            // delete the sound
      }
    }
    DOSYNC;
    s = hold;
  }
  DOSYNC;

                                   // ***********************************
                                   // *                                 *
                                   // *      Mix in fancy sounds        *
                                   // *                                 *
                                   // ***********************************
  s = soundlist;
  while(s){
    hold = s->next;
    if(s->active && s->type == v_fancy) {
      for(j = 0; j < dma_bufferlen; j++) {
                                     // Add this byte to the mixing buffer
        *(mix_buffer+j) += (*(s->data + *s->pos) * s->volume)>>4;
                                     // Here is where the frequency magic
                                     // happens.  I use fixed point arithmatic
                                     // to increment the sample position pointer
                                     // at different rates.
        s->position += s->pinc;
                                     // Are we at the end of the sample?
        if(*s->pos >= s->length) {
          if(!s->repeat) {          // Do we not repeat this sample?

            if(s->chain) {          // Is there an effect to play after this?
              (s->chain)->next = soundlist;
              soundlist = s->chain;
              s->chain = NULL;      // Break link to hide from freesound()
              sounds_in_queue++;
            }

            freesound(s->id);      // Terminate it if not.
            break;
          }
          else {                    // Decrememnt the repeat counter
            if(s->repeat > 0) s->repeat--; // -1 = infinity
            s->position = 0;
          }
        }
      }
    }
    DOSYNC;
    s = hold;
  }
  DOSYNC;

                                   // Check to make sure that dma_bufferlen
                                   // is not bigger than the number of bytes
                                   // between ticks.  If it is, this loop
                                   // will crash.
  if(mod_bytespertick < dma_bufferlen) mod_on = FALSE;

  if(mod_on && mod_data) {         // MOD stuff activated and ready?
    if(mod_reset) {                // Reset the MOD?
      for(i = 0; i < 4; i++ ) {    // Reset all four channels
        chan[i].sdata = NULL;
        chan[i].offset = 0;
        chan[i].rlength = 0;
        chan[i].end = 0;
        chan[i].volume = 0;
        chan[i].sample_number = 0;
        chan[i].position= 0;
        chan[i].pos = (WORD *)(((BYTE *)&chan[i].position) +1);
        chan[i].counter = 0;
        chan[i].pinc = 0;
        chan[i].effect = 0;
        chan[i].x = 0;
        chan[i].y = 0;
      }
                                   // Reset global mod variables
      mod_pattern=mod_data->pattern_data[mod_data->ptable[0]];
      mod_pstop = mod_pattern+1024;
      mod_tablepos=0;
      mod_bytespertick = sample_rate/11000.0 * 220;
      mod_ticksperdivision = 6;
      mod_currentbyte = 0;
      mod_currenttick = 0;
      mod_divoffset =0;

      mod_reset = FALSE;           // Reset the reset switch
    }

                                   // **************************************
                                   // This little section mixes all the mod
                                   // channels if the number of bytes left
                                   // to go until the next tick is bigger
                                   // than the dma_buffer length.
                                   // **************************************
    if(mod_currentbyte > dma_bufferlen) {
      for(i = 0; i < 4; i++) {     // Loop through the four channels
        DOSYNC;

                                   // If the volume is zero or if the sample
                                   // has ended, then there is no need to mix
                                   // anything.
        if(chan[i].volume && *chan[i].pos <= chan[i].end) {
                                   // Loop along the length of the mixing buffer
          for(j = 0; j < dma_bufferlen; j++) {
                                   // Add this byte to the mixing buffer
            *(mix_buffer+j) += (*(chan[i].sdata + *chan[i].pos) * chan[i].volume)>>mod_volume;
                                   // Here is where the frequency magic
                                   // happens.  I use fixed point arithmatic
                                   // to increment the sample position pointer
                                   // at different rates.
            chan[i].position += chan[i].pinc;
                                   // Are we at the end of the sample?
            if(*chan[i].pos >= chan[i].end) {
                                   // Repeating?  Move the position back.
              if(chan[i].rlength > 1) {
                chan[i].position = chan[i].offset*256;
                chan[i].counter = 0;
                chan[i].end = chan[i].offset+chan[i].rlength;
              }

              else break;          // else Jump out of mixing loop
            }
          }
        }
      }
                                   // Keep track of where we are.
      mod_currentbyte -= dma_bufferlen;
    }
                                   // **************************************
                                   // This little section mixes all the mod
                                   // channels if the number of bytes left
                                   // to go until the next tick is *smaller*
                                   // than the dma_buffer length.
                                   //
                                   // It first mixes what is left, then it
                                   // decrements the tick and calls out
                                   // to the module handling routines.
                                   // Once that is done, it mixes the
                                   // rest of the DMA buffer.
                                   //
                                   // The two mixing routines are extacly the
                                   // same as the one above except the first
                                   // one loops over the first part of the
                                   // mixing buffer, and the second mixes
                                   // over the rest.
                                   // **************************************

    else {
                                   // Mix the first part of the buffer
      for(i = 0; i < 4; i++) {
        DOSYNC;
        if(chan[i].volume && *chan[i].pos <= chan[i].end) {
          for(j = 0; j < mod_currentbyte; j++) {
            *(mix_buffer+j) += (*(chan[i].sdata + *chan[i].pos) * chan[i].volume)>>mod_volume;

            chan[i].position += chan[i].pinc;
                                   // Are we at the end of the sample?
            if(*chan[i].pos >= chan[i].end) {
                                   // Repeating?  Move the position back.
              if(chan[i].rlength > 1) {
                chan[i].position = chan[i].offset*256;
                chan[i].counter = 0;
                chan[i].end = chan[i].offset+chan[i].rlength;
              }
              else break;
            }
          }
        }
      }

      lastbyte = mod_currentbyte;  // Keep track of where we are inthe buffer

                                   // Ready for next division?
      if(mod_currenttick >= mod_ticksperdivision * (mod_patterndelay+1) ) {
        mod_patterndelay = 0;      // Reset pattern delay
        dodiv();                   // Handle the division commands.
        DOSYNC;
                                   // Are we at the end of the pattern?
        if(mod_pattern == mod_pstop) {
          mod_tablepos++;          // Next table!

                                   // Done with tables?  Done with MOD.
          if(mod_tablepos == mod_data->num_positions) {
            mod_on = FALSE;
            mod_reset = TRUE;

            if(mod_data->repeatmusic) { // repeat?
              if(mod_data->repeatmusic > 0) mod_data->repeatmusic--;
              mod_on = TRUE;
            }
                                   // Check for a chain function
            if(mod_data->chainfunc) mod_data->chainfunc(mod_data);

                                   // Check for a chain
            if(mod_data->chain) {
              mod_data = mod_data->chain;
              mod_on = TRUE;
            }
          }
          else {                   // Otherwise, set pointer to next table
            mod_pattern = mod_data->pattern_data[mod_data->ptable[mod_tablepos]];
            mod_pstop = mod_pattern+1024;
          }
        }
        mod_currenttick = 0;       // Reset the tick counter
      }
                                   // Reset the byte counter
      mod_currentbyte = mod_bytespertick;

      mod_currenttick++;           // Do our tick stuff.
      dotick();
      DOSYNC;
                                   // Mix the rest of the buffer
      for(i = 0; i < 4; i++) {
        DOSYNC;
        if(chan[i].volume && *chan[i].pos <= chan[i].end) {
          for(j = lastbyte; j < dma_bufferlen; j++) {
            *(mix_buffer+j) += (*(chan[i].sdata + *chan[i].pos) * chan[i].volume)>>mod_volume;

            chan[i].position += chan[i].pinc;
                                   // Are we at the end of the sample?
            if(*chan[i].pos >= chan[i].end) {
                                   // Repeating?  Move the position back.
              if(chan[i].rlength > 1) {
                chan[i].position = chan[i].offset*256;
                chan[i].counter = 0;
                chan[i].end = chan[i].offset+chan[i].rlength;
              }
              else break;
            }
          }
        }
      }
                                   // Update our position
      mod_currentbyte -= (dma_bufferlen-lastbyte);
    }

    DOSYNC;
  }

  if(debug_reverseflipflop) dma_currentbuffer ^= 1;

  DOSYNC;

                                   // Write mixed stuff to the playback buffer
  for(mbuf = mix_buffer, dbuf = dma_buffer[dma_currentbuffer];
      mbuf != mix_buffer+dma_bufferlen;
      mbuf++) {
    if(*mbuf > 254) *mbuf = 254;       // Clip overload values
    else if(*mbuf < 1) *mbuf = 1;
    *(dbuf++) = *mbuf;
  }
  DOSYNC;
  if(!debug_reverseflipflop) dma_currentbuffer ^= 1;

  if(dsp_vers < 512) {                 // Not SB 2.0?
                                          // Set up DMA for one-shot
    dmaset(dma_buffer[dma_currentbuffer],dma_bufferlen-1,dma_ch);
    DSPWrite(0x14);                       // One shot DMA DAC (8bit)
    DSPWrite((dma_bufferlen-1) & 0xff);  // Write length low byte
    DSPWrite((dma_bufferlen-1) >> 8);    // Write length high byte
  }


  DOSYNC;


                                   // Varmint's system clock
  vclock ++;                       // Vertical retrace sync clock
  vsyncclock ++;

                                   // MIDI stuff first
  midi_count-= 100;
  if(midi_count < 100) {
    midiplayer();                  // Call a midi player
    midi_count += midi_callfreq * midi_usertempo; // reset counter
    DOSYNC;
  }

  if(DSP_overhead) DSP_overhead = TimerOff();//  How long did this take?

  inportb(io_addr+DSP_RSTATUS);    // Acknowledge DSP interrupt

  outportb(0x20,0x20);              // Clear PIC
  outportb(0xa0,0x20);

  venable();                       // Enable interrupts again
}



/**************************************************************************
  void sbsetvector()

  DESCRIPTION:  Installs the DMA interrupt vector.  This makes it so that
                sbint() is called whenever a DMA transfer is finished

**************************************************************************/
void sbsetvector(void)
{
  orgirqint=_dos_getvect(int2vect(intnr));
  _dos_setvect(int2vect(intnr),handlerint);     /* set vector to our routine */
  enableint(intnr);                             /* enable sb interrupt */
}


/**************************************************************************
  void sbremovevector()

  DESCRIPTION:  Removes the DMA interrupt vector

**************************************************************************/
void sbremovevector(void)
{
  disableint(intnr);                             /* disable sb interrupt */
  _dos_setvect(int2vect(intnr),orgirqint);      /* restore org intr vector */
}


/* --------------------------------------------------- */

/* timerX routines are following

  These routines are for highly accurate time measurements

*/


/**************************************************************************
  void InitTimerFunctions()

  DESCRIPTION: Initializes speaker timer for timing operations.

**************************************************************************/
void InitTimerFunctions(void)
{
  BYTE b;

  b = inportb(0x61) & 0xfd;          // Read current port value.

                                     // Disable speaker output, but keep
                                     // timer 2 enabled.
  outportb(0x61,b | 0x01);

  measure0();                         // Setup for delay functions

}


/**************************************************************************
  void TimerOn()

  DESCRIPTION: Turns on timer counter for a time measurement

**************************************************************************/
void TimerOn(void)
{
  outportb(0x43,0xb0);                 // Set up the timer for countdown
  outportb(0x42,0xff);                // Least significant byte
  outportb(0x42,0xff);                // Most significant byte
}


/**************************************************************************
  WORD TimerOff()

  DESCRIPTION: Turns off time and reports clicks elapsed.  Note that this
               timer is so quick that it is wraps after only 56
               milliseconds.  If you want to timer longer stuff, I suggest
               using the global variable vclock.  It's tick frequency is
               sample_rate / dma_bufferlen.

               To convert to micro seconds, multiply the return value by
               0.838.

  RETURNS:
    Current timer count

**************************************************************************/
WORD TimerOff(void)
{
  outportb(0x43,0xc0);                 // Read the timer countdown status
  timer_hold = inportb(0x42) + inportb(0x42) * 256;

  timer_diff = 0xffff - timer_hold;   // timer_diff is used later
  return timer_diff;
}



/**************************************************************************
  void measure()

  DESCRIPTION: measures a standard delay loop for other delay functions

**************************************************************************/
void measure(void)
{
  vdisable();                       // Disable interrupts

  TimerOn();                        // Start the timer

  asm mov cx,10000                  /* internal test loop */
loop1:
  asm loop loop1

  timer_val = TimerOff();           // read the timer

  venable();                         // Enable interrupts

  mue999 = mcalc(990);               // Calculate a millisecond
}
/**************************************************************************
  void measure0()

  DESCRIPTION: measures a standard delay loop for other delay functions
                This is based on the standard system clock.  Use this
                functions when for some reason the timer2 functions do not
                work.

**************************************************************************/
void measure0(void)
{
  int i;
  time_t t;


  t = clock();                      // Peek at the system clock
  while(t == clock());              // Wait for system clock to tick

  for(i = 0; i < 1000; i++) {
    asm mov cx,10000                /* internal test loop */
loop1:
    asm loop loop1
  }

  t = clock() - t;                   // read the timer

  timer_val = t/1000.0 * 55000 * 1.192;  // Convert to 1.192 Mhz clock ticks

  mue999 = mcalc(990);               // Calculate a millisecond
}


/**************************************************************************
  void mdelay(WORD delay)

  DESCRIPTION: Very tiny delay

  INPUT:
    Length of delay in processor counts

**************************************************************************/
void _saveregs mdelay(WORD delay)
{
  asm mov cx,delay
loop1:
  asm loop loop1
}


/**************************************************************************
  void _saveregs MilliDelay(WORD delay)

  DESCRIPTION:  Millisec delay.  When using this library, you should
                use this delay for millisecond delays instead of the delay
                function that comes with turbo C.

  INPUTS:
    delay        Length of delay in millisconds

**************************************************************************/
void _saveregs MilliDelay(WORD delay)
{
  WORD i;

  for(i=0;i<delay;i++) mdelay(mue999); // mdelay(mue999) = 1 millisec
}

/**************************************************************************
  void _saveregs MicroDelay(WORD delay)

  DESCRIPTION:  Microsec delay.

                Note: This function probably has around a 10-15
                microsecond overhead, so I would only recommend it for
                delays bigger than 50 microseconds and smaller than 5
                milliseconds (5000 microseconds).

  INPUTS:
    delay        Length of delay in microsconds

**************************************************************************/
void _saveregs MicroDelay(WORD delay)
{
  mdelay(mcalc(delay));
}


/**************************************************************************
  WORD mcalc(WORD micro)

  DESCRIPTION:  Calculates number of ticks to send to mdelay for a specified
  number of microseconds.

  INPUTS:
    micro    Number of microseconds

  Returns:
    Value to send to mdelay() that will generate the specified number of
    microseconds.

**************************************************************************/
WORD mcalc(WORD micro)
{
  DWORD d;
  d = ((DWORD)micro*11940/timer_val);
                                      //This clip is to prevent wrap around
                                      // That can occur when a computer is
                                      // too fast or the argument
                                      // is too big.
  if(d > 0xffff) d = 0xffff;
  return((WORD)d);
}

                                      /********************************

                                                TIMER 0 STUFF




                                      *********************************/
/* --------------------------------

CAUTION: These routines can cause a lot of headaches while debugging.
If you set your own interrupt and then stop the program before you call
RemoveTimer0(), you'd better reboot your computer, because very
unpredictable things will happen if InstallTimer0() is called again.
My suggestion is to get your interrupt working and then comment out the
Timer0 routines until the rest of the program is written and debugged.
                                      - Eric  */

/**************************************************************************
  static void interrupt timerint()

  DESCRIPTION: THis is the actual interrupt function stored in the timer
               0 slot (int 08).   This calls the old int08 function
               at proper intervals as well as the user specified function

**************************************************************************/
static void interrupt timerint(void)
{

  TimerOn();                      // set timer for overhead calculation.
  call_func();                     // user specified function

                                   // Now let's do some fancy counting so
                                   // we can call the system clock at the
                                   // right moments.
  if(timsum<100)
  {
    timsum+=timadd;
    orgtick();
  }
  else
  {
    asm mov al,0x20;
    asm out 0x20,al;
  }
  timsum-=100;                     // decrement our special timer
}


/**************************************************************************
  void InstallTimer0(WORD period,void far (*func)())

  DESCRIPTION: This sets up timer0 to call your function at the specified
               period.

  INPUTS:
    period    period in system clock ticks.  (1 tick = 0.838 microsecs)
    func      pointer to the function to install

**************************************************************************/
void InstallTimer0(WORD period,void far (*func)(void))
{
   if(!func) return;       /* no valid func ptr */
  call_func=func;

  timadd= (WORD)(6553600L/period); // counting seed for timerint()
  timsum=0;                 // start counter at 0

  asm mov al,0x36            /* program timer 0 with modus 3 */
  asm out 0x43,al            /* and counter value of period  */
  asm mov ax,period
  asm out 0x40,al
  asm mov al,ah
  asm out 0x40,al

  orgtick= _dos_getvect(8);       // Remember the old interrupt
  _dos_setvect(8,timerint);      // put in a new one.
}


/**************************************************************************
  void RemoveTimer0()

  DESCRIPTION:  Removes your goofy interrupt, 'cause we didn't want
                it anyway!  :P

**************************************************************************/
void RemoveTimer0(void)
{
  if(!orgtick) return;      // Must have called InstallTimer0 first
  asm mov al,0x36            /* program timer 0 with modus 3 */
  asm out 0x43,al            /* and counter value of 0 (2^16)*/
  asm mov al,0
  asm out 0x40,al
  asm out 0x40,al
  _dos_setvect(8,orgtick);       // put back original vector
}



/**************************************************************************
  BYTE int2vect(BYTE intnr)

  DESCRIPTION:  This function converts a PIC irq number to a true
                interrupt vector number.  For PC's with a 286 or greater,
                irq's 0-7  refer to interrupts 0x08 - 0x0F and
                irq's 8-15 refer to interrupts 0x70 - 0x77.

  INPUTS:
    intnr    IRQ number

  RETURNS:
    Actual vector corresponding to the IRQ

**************************************************************************/
BYTE int2vect(BYTE intnr)
{
  if(intnr>7) return(intnr + 0x68);
  else return(intnr+8);
}


/**************************************************************************
  void enableint(BYTE nr)

  DESCRIPTION:  Enables an IRQ interrupt using the Programmable
                interrupt controller (PIC)

                To enable a interrupt, you want to turn its PIC bit off.
                Let's say that you want to turn on interrupt 3.  The
                code works something like this:


                    BIT:                 76543210

                    1 << nr             00001000
                    ~(1 << nr)          11110111 (mask)
                                        01001101 (setting)
                    setting & mask      01000101
                                            ^
                                            |
                          Notice how bit 3 is set to zero.

  INPUTS;
    nr    IRQ number

**************************************************************************/
void enableint(BYTE nr)
{
  BYTE mask,setting;

  if(nr<8) {                             // First controller?
    mask = ~(1 << nr);                  // Create interrupt mask
    setting = inportb(0x21) & mask;     // Turn off the proper bit
    outportb(0x21,setting);             // Write back the result
  }
  else {                                // Second Controller?
    mask = ~(1 << (nr-8));              // Create interrupt mask
    setting = inportb(0xa1) & mask;     // Turn off the proper bit
    outportb(0xa1,setting);             // Write back the result
  }
}
void testenableint(BYTE nr,BYTE *test)
{
  BYTE mask,setting;

  mask = ~(1 << nr);                  // Create interrupt mask
  setting = (*test) & mask;     // Turn off the proper bit
  *test = setting;             // Write back the result
}


/**************************************************************************
  void disableint(BYTE nr)

  DESCRIPTION:  Disables an IRQ interrupt using the Programmable
                interrupt controller.

                To disable a interrupt, you want to turn its PIC bit on.
                Let's say that you want to turn off interrupt 3.  The
                code works something like this:


                    BIT:                 76543210

                    1 << nr             00001000
                    ~(1 << nr)          11110111 (mask)
                                        01000101 (setting)
                    ~setting            10111010
                    (~setting) & mask    10110010 (new value of setting)
                    ~setting            01001101

                                            ^
                                            |
                          Notice how bit 3 is set to one.
  INPUTS;
    nr    IRQ number

**************************************************************************/
void disableint(BYTE nr)
{
  BYTE mask,setting;

  if(nr<8) {                            // First controller?
    mask = ~(1 << nr);                  // Create interrupt mask
    setting = (~inportb(0x21)) & mask;  // Turn on the proper bit
    outportb(0x21,~setting);            // Write back the result
  }
  else {                                // Second Controller?
    mask = ~(1 << (nr-8));              // Create interrupt mask
    setting = (~inportb(0xa1)) & mask;  // Turn on the proper bit
    outportb(0xa1,~setting);            // Write back the result
  }
}
void testdisableint(BYTE nr,BYTE *test)
{
  BYTE mask,setting;

  mask = ~(1 << nr);                  // Create interrupt mask
  setting = (~(*test)) & mask;  // Turn on the proper bit
  *test = ~setting;            // Write back the result
}


/**************************************************************************
  int getvoice(VOICE midi_voice[],int track,int channel, int note)

  DESCRIPTION: Find the first matching voice (or first inactive voice if
                a match is not found).   This function is used by midiplayer()
                as an interface to get FM voices.

  INPUTS:
    v          pointer to an array of voice structtures
    track     MIDI track that is generating this request
    channel    MIDI channel that is generating this request
    note      MIDI note requested

  RETURNS:
    Open number if successful, -1 if not.
**************************************************************************/
int getvoice(VOICE midi_voice[],int track,int channel, int note)
{
  int i;


  for(i = 0; i < 9; i++) {              // find matching active note
    if(midi_voice[i].active) {
      if(midi_voice[i].owner_track == track &&
         midi_voice[i].owner_channel == channel &&
         midi_voice[i].note == note) return(i);

    }
  }
                                        // no note, so find first inactive voice

  for(i = 0; i < 9; i++) {
    if(!midi_voice[i].active) return(i);
  }
                                        // no available voices... error
  return -1;
}

/**************************************************************************
  void dotick(void)

  DESCRIPTION:  Handles ongoing MOD commands at each tick.

**************************************************************************/
void dotick(void)
{
  int i,k;                              // Local variable we need
  WORD period,rate;
  BYTE e1,e2;
  float vibspot;

  for(i = 0; i < 4; i++) {                // Loop through all four channels
    e1 = chan[i].effect;                  // Get info aboput what is going on
    e2 = chan[i].x;

    if(chan[i].retrigger) {               // Check for a retrigger signal
                                          // Time to retrigger?
      if(!(mod_currenttick % chan[i].retrigger)) {
        chan[i].position = 0;
        chan[i].counter = 0;
        chan[i].end = mod_data->slength[chan[i].sample_number-1];
      }
    }

    if(chan[i].cut > -1) {                // Check for a cut signal
      if(mod_currenttick >= chan[i].cut) chan[i].volume = 0;
    }

    switch(e1) {
      case 0:                              // Arpeggio
        chan[i].count++;
        switch(chan[i].count % 3) {
          case 0:
            chan[i].pinc = chan[i].data1;
            break;
          case 1:
            chan[i].pinc = chan[i].data2;
            break;
          case 2:
            chan[i].pinc = chan[i].data3;
            break;
        }
        break;
      case 1:                              // Slide up
                                             // glissando = chromatic slide
        if(!mod_glissando) chan[i].period -= chan[i].porta;
        else for(k =0; k < chan[i].porta; k++) chan[i].period *= UP_SEMITONE;
                                             // Don't go too high
        if(chan[i].period < 1) chan[i].period = 1;
                                             // Set frequency counters
        rate = 7093789.2/(2*chan[i].period);
        chan[i].pinc = (rate * 256.0)/sample_rate;
        break;
      case 2:                              // slide down
        if(!mod_glissando) chan[i].period += chan[i].porta;
                                             // glissando = chromatic slide
        else for(k =0; k < chan[i].porta; k++) chan[i].period *= DOWN_SEMITONE;
                                             // Don't go too low
        if(chan[i].period > 2000) chan[i].period = 2000;
                                             // Set frequency counters
        rate = 7093789.2/(2*chan[i].period);
        chan[i].pinc = (rate * 256.0)/sample_rate;
        break;
      case 3:                              // slide to note
                                             // Slide up or down? Each case is
                                             // just like regular up and down
                                             // slides.
        if(chan[i].period < chan[i].period2) {
          if(!mod_glissando) chan[i].period += chan[i].porta;
          else for(k =0; k < chan[i].porta; k++) chan[i].period *= DOWN_SEMITONE;
          if(chan[i].period > chan[i].period2) chan[i].period = chan[i].period2;
        }
        else if(chan[i].period > chan[i].period2) {
          if(!mod_glissando) chan[i].period -= chan[i].porta;
          else for(k =0; k < chan[i].porta; k++) chan[i].period *= UP_SEMITONE;
          if(chan[i].period < chan[i].period2) chan[i].period = chan[i].period2;
        }
                                             // Set frequency counters
        rate = 7093789.2/(2*chan[i].period);
        chan[i].pinc = (rate * 256.0)/sample_rate;
        break;
      case 4:                              // vibrato
                                             // Calculate current waveform
        vibspot = chan[i].vibperiod/32.0 *
                  (double)mod_currenttick/mod_ticksperdivision * M_PI;
                                             // Adjust the period
        period = chan[i].period + chan[i].vibdepth * vibspot;
                                             // Set frequency counters
        rate = 7093789.2/(2* period);
        chan[i].pinc = (rate * 256.0)/sample_rate;

        break;
      case 5:                             // Cont. Slide to note + vol. Slide
        if(chan[i].period < chan[i].period2) {
          chan[i].period += chan[i].porta;
          if(chan[i].period > chan[i].period2) chan[i].period = chan[i].period2;
        }
        else if(chan[i].period > chan[i].period2) {
          chan[i].period -= chan[i].porta;
          if(chan[i].period < chan[i].period2) chan[i].period = chan[i].period2;
        }
        rate = 7093789.2/(2*chan[i].period);
        chan[i].pinc = (rate * 256.0)/sample_rate;

        chan[i].volume += chan[i].vslide;
        if(chan[i].volume < 0) chan[i].volume = 0;
        if(chan[i].volume > 64) chan[i].volume = 64;
        break;
      case 6:                              // Vibrato + volume slide
        vibspot = chan[i].vibperiod/32.0 *
                  (double)mod_currenttick/mod_ticksperdivision * M_PI;
        period = chan[i].period + chan[i].vibdepth * vibspot;
        rate = 7093789.2/(2* period);
        chan[i].pinc = (rate * 256.0)/sample_rate;

        chan[i].volume += chan[i].vslide;
        if(chan[i].volume < 0) chan[i].volume = 0;
        if(chan[i].volume > 64) chan[i].volume = 64;
        break;
      case 7:                              // Tremolo
                                             // Just like Vibrato except with
                                             // this modulates sample volume
       vibspot = chan[i].vibperiod/32.0 *
                  (double)mod_currenttick/mod_ticksperdivision * M_PI;
        chan[i].volume = chan[i].volume2 + chan[i].vibdepth * vibspot;
        if(chan[i].volume < 0) chan[i].volume = 0;
        if(chan[i].volume > 64) chan[i].volume = 64;
        break;
      case 10:                            // Volume slide
        chan[i].volume += chan[i].vslide;    // change the volume.
                                             // Clip.
        if(chan[i].volume < 0) chan[i].volume = 0;
        if(chan[i].volume > 64) chan[i].volume = 64;
        break;
      case 14:
        switch(e2) {
          case 13:                         //  Delay sample
                                             // data4 is decremented on each
                                             // Pass.  When it gets to zero,
                                             // the sample volume is turned
                                             // on.
            if(!chan[i].data4) chan[i].volume = chan[i].data3;
            chan[i].data4--;
            break;
          default:
            break;
        }
        break;
      case 15:
        break;
      default:
        break;
    }

    chan[i].volume *= mod_channelselect[i];  // THis allows the user to mask
                                          // channels.
  }
}
/**************************************************************************
  void dodiv(void)

  DESCRIPTION:  Handles the commands found at a module division.

**************************************************************************/
void dodiv(void)
{
  int i,j,k;                              // Local variables we'll need
  int samp,pattern_break = FALSE,position_jump = FALSE,loop = FALSE;
  WORD period,rate;
  BYTE e1,e2,e3;
  float num;

  for(i = 0; i < 4; i++) {                // Loop through all four channels
    chan[i].retrigger = 0;                // Reset retrigger and cut signals
    chan[i].cut  = -1;
                                          // Grab information for this channel
    samp = (*(mod_pattern)&0xF0) + ((*(mod_pattern+2) & 0xf0)>>4);
    period = (*(mod_pattern)&0x0F)*256 + *(mod_pattern+1);
    e1 = (*(mod_pattern+2)) & 0x0F;       // e1-e3 are the command nibbles
    e2 = (*(mod_pattern+3)  & 0xF0) >> 4;
    e3 = (*(mod_pattern+3)) & 0x0F;

    if(samp && e1 != 3) {                 // set a new sample?

      chan[i].sample_number = samp;       // Some MOD commmands need this
      chan[i].sdata = mod_data->sdata[samp-1]; // set a pointer to the data
      chan[i].position = 0;               // Retrigger the sample
      chan[i].counter = 0;
                                          // store sample characteristics
      chan[i].offset = mod_data->offset[samp-1];
      chan[i].rlength = mod_data->repeat[samp-1];
                                          // Keep track of end so we know
                                          // When to stop.
      chan[i].end = mod_data->slength[samp-1];
                                          // Set the volume to the default
      chan[i].volume = mod_data->volume[samp-1];

    }

    if(period && e1 != 3) {               // Set a new period?
                                          // Calculate adjusted period
      chan[i].period = period + mod_data->finetune[chan[i].sample_number-1]
                              + mod_finetune;
                                          // Set frequency counter
      rate = 7093789.2/(2*period);
      chan[i].pinc = (rate*256.0)/sample_rate;
    }


    chan[i].effect = e1;                   // Keep for reference in dotick()
    chan[i].x = e2;
    chan[i].y = e3;

    switch(e1) {                          // Process the commands
      case 0:                              // Arpeggio
        chan[i].data1 = chan[i].pinc;        // First frequncy

        num = chan[i].pinc;                   // calculate second frequency
        for(j = 0; j < e2; j++) num *= UP_SEMITONE;
        chan[i].data2  = (WORD)num;

        num = chan[i].pinc;                   // calculate third frequncy
        for(j = 0; j < e3; j++) num *= UP_SEMITONE;
        chan[i].data3  = (WORD)num;
        chan[i].count = 0;
        break;
      case 1:                              // Slide up.
        chan[i].porta = e2*16 + e3;          // Calculate slide rate
        break;
      case 2:                              // slide down
        chan[i].porta = e2*16 + e3;          // calculate slide rate
        break;
      case 3:                              // Slide to note
        if(e2 || e3) chan[i].porta = e2*16 + e3;
        if(period) chan[i].period2 = period;
        break;
      case 4:                              // Vibrato (My implementation
                                           // Of this command assumes that
                                           // The wave form is a triggered
                                           // Sin wave.   There are other
                                           // forms defined , but they are
                                           // use very rarely and I dont't
                                           // feel like coding them in.

                                             // These numbers will be used
                                             // to calculate the waveform.
        if(e2) chan[i].vibperiod = e2* mod_ticksperdivision;
        if(e3) chan[i].vibdepth = ((e3/16.0) * 0.0595 ) * chan[i].period;
        break;
      case 5:                              // Continue Slide to Note + Volume Slide
      case 6:                              // Continue Vibrato + volume slide
                                             // For both of these commands,
                                             // We just need to keep track
                                             // of the volume slide rate
        if((e2 > 0 && e3 > 0) || e3 == 0) chan[i].vslide = e2;
        else chan[i].vslide = -e3;
                                             // If the period is specified,
                                             // Then the volume needs to be
                                             // reset.
        if(period)chan[i].volume = mod_data->volume[chan[i].sample_number-1];
        break;
      case 7:                              // Tremolo (My implementation
                                           // Of this command assumes that
                                           // The wave form is a triggered
                                           // Sin wave.   There are other
                                           // forms defined , but they are
                                           // use very rarely and I dont't
                                           // feel like coding them in.

                                             // These values will be used to
                                             // claculate the wave form
        if(e2) chan[i].vibperiod = e2* mod_ticksperdivision;
        if(e3) chan[i].vibdepth = e3;
        chan[i].volume2 = chan[i].volume;
        break;
      case 8:                              // UNUSED
        break;
      case 9:                              // Set new sample offset
        mod_data->offset[chan[i].sample_number-1] = (e2*4096+e3*256)*2;
        break;
      case 10:                             // Volume slide
                                             // Calculate slide rate
        if((e2 > 0 && e3 > 0) || e3 == 0) chan[i].vslide = e2;
        else chan[i].vslide = -e3;
        if(period)chan[i].volume = mod_data->volume[chan[i].sample_number-1];
        break;
      case 11:                             // Table position jump
        mod_tablepos = e2*16+e3;
        position_jump = TRUE;
        break;
      case 12:                             // Set Volume
        chan[i].volume = e2*16+e3;
        break;
      case 13:                             // Pattern Break
        pattern_break = TRUE;                // We'll do the break later.
        break;
      case 14:
        switch(e2) {
          case 0:                          // Set filter.  I just ignore this
                                          // since it is a hardware option
            break;
          case 1:                          // Fineslide up
                                             // Recalculate frequency counter
            chan[i].period -= e3;
            rate = 7093789.2/(2* chan[i].period);
            chan[i].pinc = (rate*256.0)/sample_rate;
            break;
          case 2:                          // Fineslide down
                                             // Recalculate frequncy counter
            chan[i].period += e3;
            rate = 7093789.2/(2* chan[i].period);
            chan[i].pinc = (rate*256.0)/sample_rate;
            break;
          case 3:                          // Glissando on/off
                                           // If this is on, slide-to-note
                                           // functions should slide by
                                           // semitones rather than smooth
                                           // slides.  This is rarely used.
            mod_glissando = e3;
            break;
          case 4:                          // Vibrato waveform. This is rarely
                                           // used, so for now, the code is
                                           // locked in to use the default
                                           // waveform: 0  (a triggered sin
                                           // wave)
            break;
          case 5:                          // Set sample finetune value
            k = e3;
            if (k > 7) k -= 16;
            mod_data->finetune[chan[i].sample_number-1] =  k;
            break;
          case 6:                         // Loop Pattern
                                             // mod_loopcount is a global
                                             // used to keep track of the loop.
            if(mod_loopcount == -1) {
              if(!e3) mod_loopspot = mod_pattern;
              else {
                if(!mod_loopspot) mod_loopspot =
                  mod_data->pattern_data[mod_data->ptable[mod_tablepos]];
                mod_loopcount = e3;
                loop = TRUE;
              }
            }
            else if(e3){
              if(!mod_loopcount) mod_loopspot = NULL;
              else loop = TRUE;
              mod_loopcount--;
            }
            break;
          case 7:                          // Termolo waveform.  (see case #4)
            break;
          case 8:                          // UNUSED
            break;
          case 9:                         // Retrigger sample
            chan[i].retrigger = e3;          // Set the retrigger signal used
                                             // in dotick().
            break;
          case 10:                        // Fineslide volume up
            chan[i].volume += e3;
            if(chan[i].volume > 64) chan[i].volume = 64;
            break;
          case 11:                        // Fineslide volume down
            chan[i].volume -= e3;
            if(chan[i].volume < 0) chan[i].volume = 0;
            break;
          case 12:                        // Cut Sample
            chan[i].cut = e3;                // Set cut signal for dotick()
            break;
          case 13:                        // Delay sample
            chan[i].data4 = e3;              // Set counter for dotick()
            chan[i].data3 = chan[i].volume;
            chan[i].volume = 0;              // Turn off volume
            break;
          case 14:                        // Pattern Delay
            mod_patterndelay = e3;
            break;
          case 15:                        // Invert Loop.  THis is too much
                                          // of a headache to worry about
                                          // right now. Sorry.
            break;
          default:                        // Shouldn't ever get here
            break;
        }
        break;
      case 15:                            // Set Speed.
                                             // There are two possible actions
                                             // with this command:  change
                                             // ticks per beat, or change
                                             // beats per minute.  Right
                                             // now, only the ticks per beat
                                             // have been implemented.
        j = e2*16+e3;
        if(j <= 32) {
          if(!j) j = 1;
          mod_ticksperdivision = j;
        }
        break;
      default:                            // shouldn't happen.
        break;
    }
    mod_pattern += 4;                     // Advance to next channel
  }

  if(loop) {                              // Loop?  Jump back to placeholder!
    mod_pattern = mod_loopspot;
  }
  else if(pattern_break) {                // Pat. break?  Jump to next pattern!
    mod_tablepos++;
    if(mod_tablepos == mod_data->num_positions) mod_on = FALSE;
    else {
      mod_pattern = mod_data->pattern_data[mod_data->ptable[mod_tablepos]];
      mod_pstop = mod_pattern+1024;
    }
  }
  else if(position_jump) {                // Position jump?  Go to specified
                                          // Table.
    if(mod_tablepos == mod_data->num_positions) mod_on = FALSE;
    else {
      mod_pattern = mod_data->pattern_data[mod_data->ptable[mod_tablepos]];
      mod_pstop = mod_pattern+1024;
    }
  }
}


/**************************************************************************
  midiplayer()

  DESCRIPTION: Routine for playing midi files.  THis is designed to be
               called from a timer interrupt.  To use, set these values
               in this order:

                   midi_data    (must point to a filled MIDI structure.)
                  midi_reset = TRUE;
                  midi_on = TRUE;

               The interrupt should pick up from there.

               It is easy to add functionality to this routine.  I've
               already included code to flag a wide variety of MIDI
               events, so all you have to do is add your own code under
               the point an event is flagged.  I've left a bunch of
               commented print statements in to help make the code
               more readable and provide cues for accessing the data.


               ***  WARNING  ***

               If you add your own code here, make sure that it doesn't
               take more than a few milliseconds to execute.  If
               midiplayer() is called again by the interrupt before your
               code is done, your whole program will probably crash.

**************************************************************************/
void midiplayer(void)
{
  VOICE vh[9];                            // waiting list
  int vhold = 0,spot,i,j,vidx;
  BYTE event,ev,ch,b1,b2,etype;
  long int itmr,l2,length;
  char tdata[256];

  if(!midi_data) {                        // must have data to play!
    midi_on = FALSE;
    return;
  }
  if(midi_reset) {                        // Reset?  zero track pointers and timers
    for(i = 0; i < 16; i++ ) {
      midi_trackloc[i] = 1;                // no need to read first time offset
      midi_timer[i] = 0;                  // all timers start at zero
      midi_trackon[i] = TRUE;
      if(i < 9) midi_voice[i].active = 0;  // unreserve all voices
      midi_lastevent[i] = 0x80;           // set last event to note off
    }
    midi_reset = 0;                       // clear midi reset flag
    midi_livetracks = midi_data->num_tracks;// set number of active tracks so
                                          // we know when to stop.
    midi_divisions = midi_data->divisions;// ticks per quarter note
    if(midi_divisions < 0) midi_divisions = -midi_divisions;  // some midi files have
                                          // negative division values
    if(!midi_divisions) midi_divisions = 96; // Pick a default if they give us a zero
  }

  if(!midi_on) return;                     // logical switch for midi on/off


  for(i = 0 ; i < midi_data->num_tracks; i++) {  // loop over tracks
    while(midi_timer[i] <= 0) {                // Process while timer is 0;
      event = *(TD);                    // get next event (TD is a macro)
      midi_trackloc[i]++;                      // advance track location pointer

      if(event == 0xFF) {                 // META event?
        etype = *(TD);
        midi_trackloc[i] ++;
                                        // read length of meta event
        spot = readvarlen(TD,&length);
        midi_trackloc[i] += spot;
                                        // grab any text data for text events
        for(j = 0; j < length; j++) tdata[j] = *(TD + j);
        tdata[j] = 0;


        switch(etype) {
          case 0x00:
            j = *(TD)*256 + *(TD+1);
            //printf("[%d] SEQUENCE NUMBER (%d)\n",i,j);
            break;
          case 0x01:
            //printf("[%d] TEXT EVENT (%s)\n",i,tdata);
            break;
          case 0x02:
            //printf("[%d] COPYWRITE EVENT (%s)\n",i,tdata);
            break;
          case 0x03:
            //printf("[%d] TRACK NAME EVENT (%s)\n",i,tdata);
            break;
          case 0x04:
            //printf("[%d] INSTRUMENT NAME EVENT (%s)\n",i,tdata);
            break;
          case 0x05:
            //printf("[%d] LYRIC EVENT (%s)\n",i,tdata);
            break;
          case 0x06:
            //printf("[%d] MARKER EVENT (%s)\n",i,tdata);
            break;
          case 0x07:
            //printf("[%d] CUE EVENT (%s)\n",i,tdata);
            break;
           case 0x2f:                   // End of track
            //printf("[%d] END OF TRACK\n",i);
            midi_timer[i] = MAXFLOAT;               // set timer to highest value
            midi_trackon[i] = FALSE;         // turn off track
            midi_livetracks--;               // decrement track counter
            if(midi_livetracks == 0) {       // last track?  Turn off midi!
              midi_on = FALSE;
              midi_reset = TRUE;             // Make sure we start over

              if(midi_data->repeat) {       // check for repeats
                if(midi_data->repeat > 0) midi_data->repeat --;
                midi_on = TRUE;
              }
                                            // Check for a chain function
              if(midi_data->chainfunc) midi_data->chainfunc(midi_data);

                                            // Check for a chain
              if(midi_data->chain) {
                midi_data = midi_data->chain;
                midi_on = TRUE;
              }
              return;
            }
            break;
          case 0x51:                   // TEMPO event (microsecs per 1/4 note)
            l2 = *(TD) * 0x10000L + *(TD+1) * 0x100 + *(TD+2);
            //printf("[%d] TEMPO EVENT (%ld)\n",i,l2);
            midi_msperdiv = (int)(l2/midi_divisions);
                                       // Convert number to a counter used
                                       // by an 183 Hhz interrupt.
            midi_callfreq = midi_msperdiv/5454.0;
            break;
          case 0x58:
            //printf("[%d] TIME SIG EVENT (%X,%X,%X,%X)\n",i,
            //        *(TD),*(TD+1),*(TD+2),*(TD+3));
            break;
          case 0x59:
            //printf("[%d] KEY SIG EVENT (%X,%X)\n",i,*(TD),*(TD+1));
            break;
          case 0x7F:
            //printf("[%d] SEQUENCER DATA EVENT\n",i);
            break;
          default:
            //printf("[%d] *** undefined event *** (%X,type: %X,length %ld)\n",i,event,etype,length);
            break;
        }
        midi_trackloc[i] += (WORD)length;
      }
      else if(event == 0xF0 || event == 0xF7) { // sysex event
        midi_trackloc[i] += readvarlen(TD,&length);
        //printf("Sysex type 1 [length: %ld]\n",length);
        midi_trackloc[i] += (WORD)length;
      }
      else {                           // PROCESS MIDI EVENTS

        if(!(event & 0x80)) {          // top bit Not set? Running status!

          b1 = event;                  // b1 = note   (usually)
          b2 = *(TD);              // b2 = volume? (usually)


          event = midi_lastevent[i];          // use last event
          //printf("Running status >>");
          //for(j = 0; j < 9; j++) printf("%d",midi_voice[j].active);
          //printf("\n");
          midi_trackloc[i] --;                // one less byte for running status.

          if(midi_mpuout) {
            ev = event & 0xF0;
            MPUWrite(event);
            MPUWrite(b1);
            if(ev != 0xC0 && ev != 0xD0) MPUWrite(b2);
          }
        }
        else {                         // Else it was a regular event
          ch = event & 0x0f;
          ev = event & 0xF0;
          event = ev + midi_patchmap[ch];

          midi_lastevent[i] = event;          // set to last event
          b1 = *(TD);                   // get next two bytes
          b2 = *(TD+1);
          if(midi_mpuout) {
            MPUWrite(event);
            MPUWrite(b1);
            if(ev != 0xC0 && ev != 0xD0) MPUWrite(b2);
          }
        }
        ev = event & 0xF0;             // strip lower four bits
        ch = event & 0x0f;             // channel
        vidx = getvoice(midi_voice,i,ch,b1);    // Get a voice index

        switch(ev) {
          case 0x80:                   // Note off
            //printf("[%d] Note off (%d,%d)\n",i,b1,b2);
            midi_trackloc[i] += 2;
            if(vidx > -1) {            // If a matching voice was found,
                                       // kill it.
              FMKeyOff(vidx);
              FMSetVolume(vidx,0);
              midi_voice[vidx].active = FALSE;
            }
            break;
          case 0x90:                   // Note On
            //printf("[%d] Note on (%X,%d,%d)\n",i,event,b1,b2);
            //printf("%X",ch);
            midi_trackloc[i] += 2;
            if(vidx > -1) {            // Voice found?
              if(midi_voice[vidx].active) {     // already active? Turn it off.
                midi_voice[vidx].active = FALSE;
                FMKeyOff(vidx);
                FMSetVolume(vidx,0);
              }
              else if(midi_fmout){      // Wasn't active?  Turn it on.
                midi_voice[vidx].owner_track = i;
                midi_voice[vidx].owner_channel = ch;
                midi_voice[vidx].note = b1;
                midi_voice[vidx].volume = b2;
                midi_voice[vidx].active = TRUE;
                FMSetNote(vidx,b1);
                FMSetVolume(vidx,midi_volume);
                FMKeyOn(vidx);
              }
            }
            else {                     // There might be space later
                                       // store our note
              vh[vhold].owner_track = i;
              vh[vhold].owner_channel = ch;
              vh[vhold].note = b1;
              vh[vhold].volume = b2;
              vhold ++;
              if(vhold >8) vhold = 8;  //  Only nine hold notes considered
            }
            break;
          case 0xA0:                   // Key pressure
            //printf("[%d] Note presure (%d,%d)\n",i,b1,b2);
            midi_trackloc[i] += 2;
            break;
          case 0xB0:                   // Control CHange
            //printf("[%d] Control Change (%d,%d)\n",i,b1,b2);
            midi_trackloc[i] += 2;
            break;
          case 0xC0:                   // Program change
            //printf("[%d] Program change (%d)\n",i,b1);
            midi_trackloc[i] += 1;
            break;
          case 0xD0:                   // Channel Pressure
            //printf("[%d] Channel Pressure (%d,%d)\n",i,b1);
            midi_trackloc[i] += 1;
            break;
          case 0xE0:                   // Pitch wheel change
            //printf("[%d] Pitch change (%d,%d)\n",i,b1,b2);
            midi_trackloc[i] += 2;
            break;
          default:                     // Uh-OH
            //printf("MIDI ERROR (F0 midi command)\n");
            midi_on = FALSE;
            return;
        }
      }
                                         // read next time offset
      if(midi_trackon[i]) {
        midi_trackloc[i] += readvarlen(TD,&itmr);
        midi_timer[i] += itmr;
        //printf(" T: %ld\n",midi_timer[i]);
      }
    }
    midi_timer[i]-= 1.0 * midi_tempoadjust;                           // decrement timer
  }

                                        // Since there is a limited number
                                        // of FM voices, some notes do
                                        // not get voiced.  This next
                                        // section takes a list of
                                        // unallocated notes and tries
                                        // to find a spot for them.
  while(vhold) {
    vhold--;                            // go to next note
    for(i = 0; i < 9; i++) {            // loop through FM voices
      if(!midi_voice[i].active) {                // found empty one? set the note!
        midi_voice[i].owner_track = vh[vhold].owner_track;
        midi_voice[i].owner_channel = vh[vhold].owner_channel;
        midi_voice[i].note = vh[vhold].note;
        midi_voice[i].volume = vh[vhold].volume;
        midi_voice[i].active = TRUE;
        FMSetNote(i,b1);
        FMSetVolume(i,midi_volume);
        FMKeyOn(i);
        break;
      }
    }
    if(i == 9) vhold = 0;               // List full? forget about other notes
  }


}

/**************************************************************************
  MIDI far *LoadMidi(char *filename, char *errstring)

  DESCRIPTION: Reads a midi file and stores it to a MIDI data structure

  INPUTS:

    filename    Pointer to full midi filename
    errstring    Pointer to a pre-allocated string.

  RETURNS:
    returns point to MIDI structure if successful.
    NULL otherwise.

**************************************************************************/
MIDI far *LoadMidi(char *filename, char *errstring)
{
  int i;
  FILE *input;
  char sdata[256];
  long int lidata;
  int idata;
  DWORD lread;
  MIDI far *mdata;

  mdata = (MIDI far *)farmalloc(sizeof(MIDI));    // make room for music!
  if(!mdata) {
    sprintf(errstring,"Out of memory.");
    return(NULL);
  }

  mdata->chain = NULL;
  mdata->chainfunc = NULL;
  mdata->repeat = 0;

  input = fopen(filename,"rb");            // open a midi file
  if(!input) {
    sprintf(errstring,"cannot open %s",filename);
    return(NULL);
  }
                                           // Read the header

  fread(sdata,1,4,input);                  // midi id there?
  sdata[4] = 0;
  if(strcmp(sdata,"MThd")) {
    sprintf(errstring,"Not a  midi file.");
    fclose(input);
    return(NULL);
  }

  lidata = readlong(input);                // length of remaining header chunk?
  if(lidata > 250) {
    farfree(mdata);
    sprintf(errstring,"Header chunk has a weird length");
    fclose(input);
    return(NULL);
  }


  idata = readshort(input);                // Read MIDI Format specifier
  if(idata != 0 && idata != 1) {
    farfree(mdata);
    sprintf(errstring,"Unrecognized MIDI format");
    fclose(input);
    return(NULL);
  }

  mdata->format = idata;

  idata = readshort(input);                // number of tracks
  if(idata < 1 || idata > 32) {
    farfree(mdata);
    sprintf(errstring,"Bad number of tracks [%d]",idata);
    fclose(input);
    return(NULL);
  }
  mdata->num_tracks = idata;

  idata = readshort(input);                // division number (tempo)
  mdata->divisions = abs(idata);

                                           // Read individual track data
  for(i = 0; i < mdata->num_tracks; i++) {
    fread(sdata,1,4,input);                // midi track id there?
    sdata[4] = 0;
    if(strcmp(sdata,"MTrk")) {
      farfree(mdata);
      sprintf(errstring,"Error reading track #%d",i);
      fclose(input);
      return(NULL);
    }

    lidata = readlong(input);              // length of remaining track chunk?

                                           // Allocate space for track
    mdata->track[i] = (BYTE far *)farmalloc(lidata);
    if(!mdata->track[i]) {
      while(i) {
        i--;
        farfree(mdata->track[i]);
      }
      farfree(mdata);
      sprintf(errstring,"Out of memory.");
      fclose(input);
      return(NULL);
    }
                                           // read in entire track
    lread = fread(mdata->track[i],1,(size_t)lidata,input);
    if(lread < lidata) {
      while(i) {
        i--;
        farfree(mdata->track[i]);
      }
      farfree(mdata);
      sprintf(errstring,"Premature end of midi file [track %d]",i);
      fclose(input);
      return(NULL);
    }
  }

  fclose(input);
  return mdata;
}


/**************************************************************************
   int readvarlen(char *data,long int *value)

  DESCRIPTION:  Reads a variable length long interger from data string.
                This functionis used by LoadMidi() to grab special numbers
                from MIDI data files.

  INPUTS:
    data    pointer to a character string that we want to convert
    value    pointer to variable where we will diump the value

  RETURNS:
    Number of bytes read.

**************************************************************************/
int readvarlen(BYTE *data,long int *value)
{
  int i=0;
  BYTE c;

  if ((*value = *(data + i)) & 0x80) { // Process the variable length interger
    *value &= 0x7f;
    do {
      i++;
      *value = (*value << 7) + ((c = *(data +i)) & 0x7f);
    } while (c & 0x80);
  }
  return(i+1);                          // return number of bytes read
}


/**************************************************************************
  long int readshort(FILE *inflile)

  DESCRIPTION:  Reads a short interger from a file

  INPUTS:
    infile    pointer to an open file stream

  RETURNS:
    value of interger read

**************************************************************************/
int readshort(FILE *infile)
{
  return (fgetc(infile) << 8) | fgetc(infile);
}

/**************************************************************************
  long int readlong(FILE *inflile)

  DESCRIPTION:  Reads a long interger from a file

  INPUTS:
    infile    pointer to an open file stream

  RETURNS:
    value of interger read

**************************************************************************/
long int readlong(FILE *infile)
{
  int i;
  long int num = 0;

  num = (unsigned char)fgetc(infile);

  for(i = 0; i < 3; i++) {
    num = (num << 8) | (unsigned char)fgetc(infile);
  }

  return(num);
}



/**************************************************************************
  void freesound(DWORD id)

  DESCRIPTION:  Removes sound effect from active list and clears token for
                later use.

  INPUTS:
    id    Id of sound effect to free

**************************************************************************/
void freesound(DWORD id)
{
  SOUNDEFFECT far *s,*hold,*hold2;


  hold = NULL;                       // Find the sound based on the id
  s = soundlist;
  while(s && s->id != id) {          // move down until we find the id
    hold = s;
    s = s->next;
  }

  if(!s) return;                     // No such sound? go home!

  sounds_in_queue--;                  // keep track for PlaySound()
  s->type = v_unused;
  s->id = 0;

  if(!hold) {                        // Hold empty? Must be head of list
    soundlist = s->next;             // Close the list
    return;
  }

  hold->next = s->next;              // close the spot in the list

  hold = s->chain;                   // Kill any chains attached to it
  while(hold) {
    hold2 = hold->chain;
    hold->type = v_unused;
    hold->id = 0;
    hold = hold2;
  }
  return;
}


/**************************************************************************
  SOUNDEFFECT far *newsoundeffect(void)

  DESCRIPTION:  Creates a new sound effect an initializes it variables

                Note: Soundeffect Tokens are pre-allocated rather than dynamically
                      allocated because many thousands of sounds may
                      be played in a single game.  Allocating and
                      re-allocating all this tiny pieces of memory can cause
                      problems.

  RETURNS:
    Pointer to the new sound effect structure

   SAMPLE   *data;
  DWORD   length;
  DWORD   position;
  WORD     *pos;
  WORD     pinc;
  int     repeat;
  BYTE     volume;
  BYTE    active;
  DWORD   id;
  struct soundeffect *chain;
  struct soundeffect *next;

**************************************************************************/
SOUNDEFFECT far *newsoundeffect(void)
{
   SOUNDEFFECT far *s;
  int i;
                                   // Find an empty sound token
  for(i =0 ; i< MAXSOUNDTOKENS && soundtoken[i].type != v_unused; i++);
  if(i == MAXSOUNDTOKENS) return(NULL);
  s = &soundtoken[i];
                                   // Initialize varaibles
  s->data = NULL;
  s->length = 1;
  s->position = 0;
  s->pos = (WORD *)(((BYTE *)&(s->position))+ 1);
  s->pinc = 256;
  s->repeat = 0;
  s->volume = 16;
  s->active = TRUE;
  s->id = sampleid++;
  s->chain = NULL;
  s->next = NULL;
  s->chainfunc = NULL;
  return(s);
}

/**************************************************************************
  DWORD PlaySound(SAMPLE *data,DWORD length,VATPLAYTYPE type)

  DESCRIPTION: Adds an sound to the play list.  If the playlist is full,
               the oldest one gets chopped;

  INPUTS:
    data      pointer to sample data
    length    length of sample in bytes
    type      inidicates which code is used to play the sound

  RETURNS:
    id of created sound effect or NULL on failure.  This id can is used
    as a handle to access the effect by other functions.


**************************************************************************/
DWORD PlaySound(SAMPLE *data,DWORD length,VATPLAYTYPE type)
{
  SOUNDEFFECT far *s;

  _disable();                               // We don't want any interrupts
                                            // While we do this.

  if(sounds_in_queue >= MAXSOUNDS) {          // Too many sounds active?
                                            // Chop off sound at end of list
    s = soundlist;
    while(s && s->next) s = s->next;
    freesound(s->id);                       // Kill the sound
  }

  s = newsoundeffect();                      // Allocate new effect
  if(!s) {
    _enable();                              // restore interrupts
    return(0);
  }

  s->data = data;                            // Initialize data portion
  s->length = length-1;
  if(type == v_plain) *s->pos = (WORD)s->length;
  s->type = type;

  s->next = soundlist;                        // Add to the list
  soundlist = s;

  sounds_in_queue++;

  _enable();                                // retore interrupts
  return(s->id);
}


/**************************************************************************
  DWORD ChainSoundEffect(SAMPLE *data,length)

  DESCRIPTION:  This chains a new soundeffect to an active one.  The new
                sound effect wiill be played when the first one os finished.

  INPUTS:
    data      pointer to sample data
    length    length of sample in bytes
    type      playtype for the sew sample (v_fancy or v_plain)
    id        id if the sound effect to chain to

  RETURNS:
    id of created sound effect or NULL on failure.  This id can is used
    as a handle to access the effect by other functions.


**************************************************************************/
DWORD ChainSoundEffect(SAMPLE *data,DWORD length,VATPLAYTYPE type,DWORD id)
{
  SOUNDEFFECT far *s,*o;

  _disable();                               // We don't want any interrupts
                                            // While we do this.

  o = grabsound(id);                       // grab the original effect
  if(!o) {
    _enable();                             // Restore interrupts
    return(0);
  }

  s = newsoundeffect();                    // Allocate new effect
  if(!s) {
    _enable();                             // restore interrupts
    return(0);
  }

  s->data = data;                          // Initialize data portion
  s->length = length-1;
  if(type == v_plain) *s->pos = (WORD)s->length;
  s->type = type;
                                           // Link the new soundeffect
  while(o->chain) o = o->chain;            // Find the end of the chain
  o->chain = s;

  _enable();                               // Restore interrupts
  return(s->id);
}


/**************************************************************************
  int AddChainFunction(DWORD id,void (*func)(SOUNDEFFECT far *))

  DESCRIPTION:  Attaches a function to a sound effect.  The function is
                called when the sound effect is done.

  INPUTS:
    id      Id of sound effect to use
    func    pointer to the function to chain to

  RETURNS:
    TRUE if successful, FALSE if not

**************************************************************************/
int AddChainFunction(DWORD id,void (*func)(SOUNDEFFECT far *))
{
  SOUNDEFFECT far *s;

  s = grabsound(id);              // find the soundeffect
  if(!s) return(FALSE);

  s->chainfunc = func;            // add the function

  return(TRUE);                   // It worked!
}

/**************************************************************************
  SOUNDEFFECT far *grabsound(DWORD id)

  DESCRIPTION:  Finds a sound structure matching the id

  INPUTS:
    id      id of the sound structure to find

**************************************************************************/
SOUNDEFFECT far *grabsound(DWORD id)
{
  int i;
                                // Search sound tokens for matching id
  for(i = 0; i < MAXSOUNDTOKENS && soundtoken[i].id != id; i++);
  if(i == MAXSOUNDTOKENS) return NULL;

  return(&soundtoken[i]);       // Return what we got
}

/**************************************************************************
  void AlterSoundEffect(DWORD id,VATCOMMAND c,long int parameter)

  DESCRIPTION:  Alters items in the sound effect structure

  INPUTS:
    id          soundeffect id
    c            The command id to signal the appropriate action
    parameter   The argument of the command

**************************************************************************/
void AlterSoundEffect(DWORD id,VATCOMMAND c,double parameter)
{
  SOUNDEFFECT far *s;

  s = grabsound(id);

  if(!s) return;

  switch(c) {                            // Act on all the varmint commands
    case v_play:
      s->position = 0;
      s->active = TRUE;
      break;
    case v_stop:
      freesound(id);
      break;
    case v_pause:
      s->active = FALSE;
      break;
    case v_resume:
      s->active = TRUE;
      break;
    case v_rewind:
      s->position = 0;
      break;
    case v_setrepeat:
      s->repeat = parameter;              // positive value = loop x times
                                          // negative value = loop forever
                                          // 0 = loop once
      break;
    case v_setpos:
      *s->pos = parameter;
      break;
    case v_setrate:                       // Rate is sent in as a ratio
      s->pinc = 256 * parameter;
      break;
    case v_setvolume:
      s->volume = parameter;              // 32 = LOUD, 0 = silent
      break;
    default:
      break;
  }

}


/**************************************************************************
  void ModCommand(VATCOMMAND c)

  DESCRIPTION:  Simple Interface for working With MOD playback

  INPUTS:
    c            VAT command (See sound.h)

**************************************************************************/
void ModCommand(VATCOMMAND c)
{
  switch(c){
    case v_play:
      mod_reset = TRUE;
      mod_on = TRUE;
      break;
    case v_stop:
      mod_on = FALSE;
      mod_reset = TRUE;
      break;
    case v_pause:
      mod_on = FALSE;
      break;
    case v_resume:
      mod_on = TRUE;
      break;
    case v_rewind:
      mod_reset = TRUE;
      break;
    default:
      break;
  }
}

/**************************************************************************
  void MidiCommand(VATCOMMAND c)

  DESCRIPTION:  Simple Interface for working With Midi playback

  INPUTS:
    c            VAT command (See sound.h)

**************************************************************************/
void MidiCommand(VATCOMMAND c)
{
  switch(c){
    case v_play:
      midi_reset = TRUE;
      midi_on = TRUE;
      break;
    case v_stop:
      midi_on = FALSE;
      midi_reset = TRUE;
      break;
    case v_pause:
      midi_on = FALSE;
      break;
    case v_resume:
      midi_on = TRUE;
      break;
    case v_rewind:
      midi_reset = TRUE;
      break;
    default:
      break;
  }
}

/**************************************************************************
  MOD far  *LoadMod(char *filename,char *errstring)

  DESCRIPTION:  Loads a Module file (Noise,Sound, or Protracker)

  INPUTS:
    filename    pointer to character string containing the MOD filename
    errstring   pointer to a string for dumping error output

  RETURNS:
    pointer to a new MOD structure filled with MOD data

**************************************************************************/
MOD far *LoadMod(char *filename,char *errstring)
{
  int i,j,ft;
  FILE *input;
  MOD far *m;
  char signature[5];
  int samples = 15;

  m = (MOD far *)farmalloc(sizeof(MOD));         // Allocate the skeleton MOD struct
  if(!m) {
    sprintf(errstring,"Out of memory");
    return(NULL);
  }

  m->chain = NULL;
  m->chainfunc = NULL;
  m->repeatmusic = 0;

  input = fopen(filename,"rb");           // access the file
  if(!input) {
    farfree(m);
    sprintf(errstring,"Cannot open file %s",filename);
    return(NULL);
  }

  fseek(input,1080, SEEK_SET);            // Look for signatures so we
                                          // can tell of this MOD has 15
                                          // Or 31 samples.
  fread(signature,1,4,input);
  signature[4] = 0;
  if(!strcmp(signature,"M.K.")) samples = 31;
  else if(!strcmp(signature,"M!K!")) samples = 31;
  else if(!strcmp(signature,"FLT4")) samples = 31;

  rewind(input);                          // Go back to the start.

  fread(m->title,1,20,input);              // read the module title

  m->num_samples = 0;

  for(i = 0; i < samples; i++) {          // read the sample headers
    fread(m->sample_name[i],1,22,input);  // name
                                          // length (convert words to bytes)
    m->slength[i] = (fgetc(input) * 256+ fgetc(input))*2;
    ft = fgetc(input)&0x0f;                // finetune
    if(ft>7) ft -=16;                     // Convert to a signed value
    m->finetune[i] = ft;

    m->volume[i] = fgetc(input);          // Sample volume
                                          // repeat offset (words -> bytes)
    m->offset[i] = (fgetc(input) * 256 + fgetc(input))*2;
                                          // repeat length (words -> bytes)
    m->repeat[i] = (fgetc(input) * 256 + fgetc(input))*2;
    if(m->repeat[i] == 2) m->repeat[i] = 0;

                                          // Sometimes the repeat length
                                          // and offset are not set properly,
                                          // so we have to do a little
                                          // patching here.
    if(m->offset[i] + m->repeat[i] > m->slength[i] ) {
      if(m->repeat[i] > m->slength[i]) {
        m->repeat[i] = m->slength[i];
        m->offset[i] = 0;
      }
      else {
        m->offset[i] -= (m->repeat[i] + m->offset[i] - m->slength[i]);
      }
    }

    if(m->slength) m->num_samples++;      // Keep track of the number of real smaples
  }

  m->num_positions = fgetc(input);        // number of patterns to play
  fgetc(input);                           // unused byte
  fread(m->ptable,1,128,input);           // read pattern table

  if(samples == 31) {
    fread(m->sig,1,4,input);              // read module signature
    m->sig[4] = 0;
  }
  else strcpy(m->sig,"NONE");

  m->maxpattern = 0;

  for(i = 0; i < m->num_positions; i++) { // Find number of unique patterns
    if(m->ptable[i] > m->maxpattern) m->maxpattern = m->ptable[i];
  }

  for(i=0;i<128;i++) m->pattern_data[i]=NULL;  // Reset pattern buffer pointers

  for(i = 0; i <= m->maxpattern; i++) {   // read in 1024-byte patterns
    m->pattern_data[i] = (BYTE far *)farmalloc(1024);

    if(!m->pattern_data[i]) {             // Did we get the memory?
                                          // If not, free what we have
                                          // and go home.
      for(j = 0; j < i; j++) farfree(m->pattern_data[j]);
      sprintf(errstring,"Out of memory");
      farfree(m);
      return(NULL);
    }
    fread(m->pattern_data[i],1,1024,input);  // Read in the pattern
  }
                                          // Reset sample data pointers
  for(i = 0; i < 32; i++) m->sdata[i] = NULL;

  for(i = 0; i <samples; i++) {           // Read in the sample data
    m->sdata[i] = NULL;                   // Reset sample
    if(!m->slength[i]) continue;          // Continue if nothing to read

    m->sdata[i] = (SAMPLE *)farmalloc(m->slength[i]+1);

    if(!m->sdata[i]) {                    // Did we get the memory?
                                          // If not, free what we have
                                          // and go home.
      for(j = 0; j <= m->maxpattern; j++) farfree(m->pattern_data[j]);
      for(j = 0; j < i; j++) farfree(m->sdata[j]);
      sprintf(errstring,"Out of memory");
      farfree(m);
      return(NULL);
    }
                                          // Read in the sample
    fread(m->sdata[i],1,m->slength[i],input);

  }
  fclose(input);                          // All done!


  return(m);                              // Give the new Module back to
                                          // the caller.

}



/*********************************************************************
  SAMPLE *LoadWave(char *wavefile,unsigned long int *length)

  DESCRIPTION:  Loads a wave files  (mono, 8bit)

  INTPUTS:

    wavefile  filename of wave file
    length    pointer to length interger

  RETURNS:
    pointer to data

**********************************************************************/
SAMPLE *LoadWave(char *wavefile,unsigned long int *length)
{
  int i;
  SAMPLE  *data;
  BYTE dummydata[255];
  FILE *input;
  DWORD rlen,flen;
  WORD s_per_sec,b_per_sec,num_channels,tag;
  char riffid[5],waveid[5],fmtid[5],dataid[5];

  input = fopen(wavefile,"rb");
  if(!input) {                              // If unsuccesful...
    *length = 1;                             // set short length to prevent
                                            // mistakes later.
    return(NULL);                           // REturn a null pointer
  }
                                            // Get WAVE header data

  fread(riffid,1,4,input);                  // wave files staqrt with "Riff"
  riffid[4] = 0;
  fread(&rlen,1,4,input);                   // File size
  fread(waveid,1,4,input);                  // Wave id string  ("Wave")
  waveid[4] = 0;
  if(strcmp(waveid,"WAVE")) {               // is it a wave file?
    fclose(input);
    return(NULL);
  }

  fread(fmtid,1,4,input);                   // Format id string ("fmt ")
  fmtid[4] = 0;
  fread(&flen,1,4,input);                   // offset to data
  if(flen > 240) flen = 240;                // Just a precaution so that
                                            // We do not overload dummydata

  fread(&tag,1,2,input);                    // tag
  fread(&num_channels,1,2,input);           // number of channels
  fread(&s_per_sec,1,2,input);              // sample rate (hz)
  fread(&b_per_sec,1,2,input);              // bytes per seconf rate
  fread(dummydata,1,(size_t)flen-8,input);          // Skip ahead
  fread(dataid,1,4,input);                  // Dataid string
  dataid[4] = 0;
  fread(length,1,4,input);                  // length of data

  data = (SAMPLE *)farmalloc(*length+1);    // allocate memory for data
  if(!data) {                                // oops.  Not enough mem!
    fclose(input);
    return(NULL);
  }

  fread(data,1,(size_t)*length,input);              // read the data

  for(i = 0; i < *length; i++) {            // convert to signed format
    *(data + i) = ((BYTE)*(data + i))-128;
  }

  fclose(input);                            // Wrap it up
  return(data);
}


/**************************************************************************
  void LoadInstruments(char *filename,BYTE inst[128][11])

  DESCRIPTION:  Loads instrument defs from a file (128 total)

    File format: 11 hex values followed by a name.  eg:

      30 33 40 00 E1 E2 87 63 06 01 00 "Electric Piano 2"
      33 34 00 00 92 C3 C3 B3 02 01 00 "Harpsichord"
      32 32 00 00 92 C3 C3 B3 02 01 00 "Clavichord"
      .
      .
      .

    (The text at the end is ignored.)

    The hex values are dumped into an 2-D array.  The file can have
    more or less than 128 defs without harm to this function.

  INPUTS:
    filename    pointer to a string with the instrument file name in it
    inst        pointer to a 2D instrument array (should be [128][11])

**************************************************************************/
int LoadInstruments(char *filename,BYTE inst[128][11])
{
  FILE *input;
  int i=0,j;
  char string[255];

  input = fopen(filename,"r");     // open the file
  if(!input) return(0);
                                   // read  it's contents
  while(fgets(string,255,input) && i < 128) {
    for(j = 0; j < 11; j++) sscanf(string+j*3,"%X ",&inst[i][j]);
    i++;
  }
                                   // clean up and go home
  fclose(input);
  return(i);
}



/**************************************************************************
  void FreeMidi(MIDI *m)

  DESCRIPTION: Frees the data allocated for a MIDI structure

  INPUTS:
    m    pointer to MIDI structure to free up

**************************************************************************/
void FreeMidi(MIDI *m)
{
  int i;

  for(i = 0; i < m->num_tracks; i++) free(m->track[i]);
  farfree(m);
}

/**************************************************************************
  void FreeMod(MOD *freeme)

  DESCRIPTION: Frees all the memory associated with a MOD file.

  INPUTS:
    freeme    pointer to MOD structure to free up

**************************************************************************/
void FreeMod(MOD *freeme)
{
  int i;

  for(i = 0; i <32; i++) {           // Free the samples
    if(freeme->sdata[i]) farfree(freeme->sdata[i]);
  }
  for(i = 0; i <128; i++) {         // Free the pattern data
    if(freeme->pattern_data[i]) farfree(freeme->pattern_data[i]);
  }
  free(freeme);

}

/**************************************************************************
  void VarmintVSync(void)

  DESCRIPTION: This vsync function pays attention to vsyncclock so that we
                 never miss a vsync.  The vsync can be missed if sbint
                gets called while we are waiting for the retrace.

                The variable vsync_toolong should be roughly equalt to
                the number of times you expect sbint to get called between
                vertical retraces.  The will be afffected by dma_bufferlen,
                sample_rate, and the frequency of the VGA monitor

**************************************************************************/
void VarmintVSync(void)
{
  while(!(inportb(0x3da)&0x08)) {     // Wait for retrace to start
    if(vsyncclock>vsync_toolong) break;   // Have we waited too long?
  }

  while(inportb(0x3da)&0x08);          // Wait for retrace to finish
  vsyncclock = 0;                     // reset the clock
}


/**************************************************************************
  void TimeVSync(void)

  DESCRIPTION:  This times a vertical retrace and sets the variable
                vsynctoolong to an appropriate value based on the sample
                rate and dma_bufferlen.

                This should be called before GoVarmint();

**************************************************************************/
void TimeVSync(void)
{
  WORD l;

  VarmintVSync();                // Wait for a Vync to end

  TimerOn();                    // Time the next vsync
  VarmintVSync();
  l = TimerOff();

  if(l < 100) l = 15000;         // Sanity measure

                                 // Figure how many vclock ticks will occur
                                 // During a retrace.
  vsync_toolong = l  / 1193000.0 * sample_rate/dma_bufferlen + 1.5;
}


/**************************************************************************
  float PercentOverhead(WORD ticks)

  DESCRIPTION:  Calculates the Percentage CPU overhead that the
                blaster interrupt consumes.

  INPUTS:
    ticks        Number of clock ticks that pass during the interrupt

  RETURNS:
    A percentage based on dma_bufferlen and sample rate


**************************************************************************/
double PercentOverhead(WORD ticks)
{
  return ((ticks / 1193000.0)* (sample_rate / dma_bufferlen) * 100.0);
}


/**************************************************************************
  void vdisable(void)

  DESCRIPTION:  Special interrupt disable function for VAT

**************************************************************************/
void vdisable(void)
{
  if(disabled) {                 // Don't do anything if interrupts are
    disabled++;                  // already disabled.
    return;
  }

  _disable();                    // disable the interrupts
  disabled++;
}

/**************************************************************************
  void venable(void)

  DESCRIPTION:  Special interrupt enable function for VAT

**************************************************************************/
void venable(void)
{
  disabled--;

  if(!disabled) _enable();       // enable the interrupts if this is the
                                 // outermost enable.
}