/**************************************************************************
                        VARMINT'S AUDIO TOOLS

  Header file: SOUND.H

  Written by: Peter Sprenger

  Modifications by:  Eric Jorgensen (Feb, 1995)
                    smeagol@rt66.com


 Peter Sprenger's Original Copywrite is as follows:


 * Copyright 1993 by Peter Sprenger   Pete@amber.dinoco.de
 *                   5014 Kerpen 3
 *                   Germany
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The author Peter Sprenger
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.

 In the spirit of Peter's effort, I am offering my modification of his
 library as FREEWARE.  You are free to use it and distribute it as long
 as no fee is charged.

                           **** WARNING ****

 Use Varmint's Audio tools at your own risk.  This code has not undergone
 any sort of rigourous testing and has been found to cause cancer in
 laboratory rats.

**************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <string.h>
#include <ctype.h>
#include "math.h"
                               // More definitions than you can
                               // shake a stick at.
#define VAT_VERSION "0.51"
                               // Pointer type definition for samples.
                               // This definition should be far or
                               // huge.  If you use near, you will have
                               // to change the code to use malloc()
                               // instead of farmalloc()
#define SAMPLE signed char

#define MAXSOUNDS 16

#define WORD unsigned short
#define DWORD unsigned int
#define BYTE unsigned char
#define FALSE 0
#define TRUE 1

#define DSP_RESET 6
#define DSP_READ 0x0a
#define DSP_WRITE 0x0c
#define DSP_WSTATUS 0x0c
#define DSP_RSTATUS 0x0e

#define FM_DETECT 1
#define DSP_DETECT 2

#define DSP_DMA_8BIT 0x14
#define DSP_DMA_ADC 0x24
#define DSP_SAMPLE_RATE 0x40

#define DSP_MIDI_WRITE 0x38
#define DSP_MIDI_READI 0x31
#define DSP_MIDI_UART 0x35
#define DSP_SILENCE 0x80
#define DSP_SPKR_ON 0xd1
#define DSP_SPKR_OFF 0xd3
#define DSP_GET_VERS 0xe1

#define DSP_INVOKE_INTR 0xf2

//#define FM_BOTH_OFF  0x168
#define FM_BOTH_OFF  0x08
#define FM_LEFT_OFF  0
#define FM_RIGHT_OFF 2

#define FM_HIHAT     1
#define FM_TOPCYM    2
#define FM_TOMTOM    4
#define FM_SNARE     8
#define FM_BASS   0x10


#define M_PI        3.14159265358979323846
#define MAXFLOAT    3.37E+38


typedef enum {bit8=0,bit4=1,bit26=2,bit2=3,high} playmode;
typedef enum {none,nodsp,adlib,sb20,sbpro} cardtype;
typedef enum {both,left,right} fm_chan;
typedef enum {midi,sample} vect;
typedef enum {fmerr=0,nodsperr,irqerr,dmacherr,nomem} SBERROR;

typedef struct {
  SAMPLE *sdata;
  WORD offset,rlength,end;
  signed char volume,volume2;
  WORD sample_number;
  WORD position;
  WORD counter,pinc;
  BYTE effect,x,y;
  short data1,data2,data3,data4;
  WORD count;
  short porta;
  BYTE retrigger;
  signed char cut;
  WORD period,period2,vibdepth,vibperiod;
  short vslide;
} CHANNEL;

typedef struct {
  char title[20];
  char sample_name[32][22];
  WORD slength[32],offset[32],repeat[32];
  signed char finetune[32];
  BYTE volume[32],num_positions,maxpattern,ptable[128],sig[5];
  BYTE *pattern_data[128];
  BYTE num_samples;
  SAMPLE *sdata[32];
} MOD;

typedef struct {                      // MIDI Voice data structure
  short owner_track;
  short owner_channel;
  short active;
  short note;
  short volume;
} VOICE;


typedef struct {                      // MIDI data structure
  short format;                         // 0 or 1
  short num_tracks;                     // 1 - 16
  short divisions;                       // number ticks per quarter note
  BYTE *track[16];                    // Track data pointers
} MIDI;

void FM_Write(WORD data);
void FM_Reset(void);
BYTE FM_Status(void);
void FM_SetChan(fm_chan channel);
void FM_SetVoice(BYTE voice,BYTE *ins);
void FM_SetNote(BYTE voice,BYTE note);
void FM_SetFreq(BYTE voice,WORD freq);
void FM_KeyOn(BYTE voice);
void FM_KeyOff(BYTE voice);
void FM_SetVol(BYTE voice,BYTE vol);
void FM_RythmMode(BYTE bool);
void FM_RythmOn(BYTE inst);
void FM_RythmOff(BYTE inst);

void SetRate(WORD rate);

short SB_Setup(void);
                                   // These functions Are not really needed
                                   // outside of sound.c

//short DSP_Reset();
//BYTE DSP_Read();
//void DSP_Write(BYTE output);
//WORD DSP_Get_Version();
//void SB_SetVect();
//void SB_RemoveVect();
//short get_sb_env();
//short CardCheck();
//cardtype WhichCard();
//BYTE int2vect(BYTE intnr);
//void enable_int(BYTE nr);
//void disable_int(BYTE nr);
//void InitT2();
//void measure();
//void dma_set(BYTE far *sound_address,WORD len,BYTE channel);
//WORD polldma(BYTE channel);
//short ReadVarLen(BYTE *data,int *value);
//int ReadLong(FILE *infile);
//short ReadShort(FILE *infile);


/* timerx prototypes */

void timer_on(void);
WORD timer_off(void);
WORD to_micro(WORD clk);
void clkdelay(WORD clicks);
void mdelay(WORD delay);
WORD mcalc(WORD micro);
void Install_Timer0(WORD period,void (*func)(void));
void Remove_Timer0();
void _saveregs msdelay(WORD delay);


/*  Stuff added by Eric Jorgensen */
					  // function prototypes
short     ReadMidi(char *filename, MIDI **midipoint, char *errstring);
void     MidiPlayer(void);
void     playsound(SAMPLE *data,DWORD length);
SAMPLE   *loadwave(char *wavefile,unsigned int *length);
void     Go_Varmint(void);
void     Dropdead_Varmint(void);
short     load_instruments(char *filename,BYTE inst[128][11]);
void     freemidi(MIDI *m);
short     MPU_Reset(void);
void     MPU_Write(BYTE b);
MOD     *loadmod(char *modfilename);
void     freemod(MOD *freeme);
void     VarmintVSync(void);

/* Function added for WATCOM C port by Brian Gavin */
void     real_free(int handle);
void *real_malloc(int size,int* handle);

					  // Globals
					  // The variables athat are
                                          // commented out should
                                          // probably stay that way since
                                          // They are only used internally.
extern MIDI *midi_data;
extern short midi_reset;
extern short midi_on;
extern WORD midi_port;
extern short midi_dumbmode;
extern float midi_tempoadjust;
extern float midi_usertempo;
extern BYTE music_volume;
extern char *errname[];
extern DWORD vclock;
extern SBERROR sberr;
extern WORD DSP_overhead;
extern short sounds_in_queue;
extern WORD dma_bufferlen;
//extern WORD dma_bytesahead;
extern BYTE *dma_buffer;
extern BYTE *playahead_buffer;
extern WORD intnr,card_id;
extern int dma_ch;
extern short debugnum;
extern float midi_callfreq;
extern WORD io_addr;
extern SAMPLE *sounddata[MAXSOUNDS];
extern DWORD soundpos[MAXSOUNDS];
extern short sample_rate;               // FYI. Do not change this number.
extern void (interrupt *handlerint)(void);
extern short *mix_buffer;
extern short mpu_timeout;
extern short mpu_available;
extern MOD *mod_data;
extern short mod_on,mod_reset,mod_tablepos,mod_currenttick;
extern short mod_bytespertick ,mod_ticksperdivision;
extern WORD vsyncclock,vsync_toolong;
extern CHANNEL chan[4];
extern BYTE channel_select[4];
extern BYTE sync_on;
extern short mod_volume;
extern short midi_mpuout;
extern short midi_fmout;
extern BYTE *midi_patchmap;
extern WORD dsp_vers;
extern short debug_antistatic;
extern short debug_intdisable;
